# A thread that manages all Weave IO (we haven't migrated everything here yet).
# You can perform IO by using the get_sync_client or get_async_client interfaces.

# Warning: if you see errors coming from here, logging in the server process
# is not currently working. You can change logger.error to print to get stack
# traces in local development.
# TODO: Fix

import time
import atexit
import queue
import uuid
import asyncio
import dataclasses
import typing
import contextlib
import aioprocessing
import multiprocessing
import logging
import traceback
import threading


from . import artifact_wandb
from . import errors
from . import engine_trace
from . import filesystem
from . import weave_http
from . import wandb_api
from . import wandb_file_manager
from . import server_error_handling
from . import async_queue
from typing import Any, Callable, Dict, TypeVar, Iterator


tracer = engine_trace.tracer()  # type: ignore
statsd = engine_trace.statsd()  # type: ignore


QueueItemType = TypeVar("QueueItemType")


class ArtifactMetadata(typing.TypedDict):
    created_at: str


# ServerRequestContext holds the context for server requests
@dataclasses.dataclass
class ServerRequestContext:
    trace_context: typing.Optional[engine_trace.TraceContext]
    wandb_api_context: typing.Optional[wandb_api.WandbApiContext]

    @classmethod
    def from_json(cls, json: typing.Any) -> "ServerRequestContext":
        trace_context = engine_trace.new_trace_context()
        if trace_context:
            trace_context.__setstate__(json["trace_context"])
        wandb_api_context = None
        wandb_api_context_json = json.get("wandb_api_context")
        if wandb_api_context_json:
            wandb_api_context = wandb_api.WandbApiContext.from_json(
                wandb_api_context_json
            )
        return cls(trace_context=trace_context, wandb_api_context=wandb_api_context)

    def to_json(self) -> typing.Any:
        trace_context = None
        if self.trace_context:
            trace_context = self.trace_context.__getstate__()
        wandb_ctx = None
        if self.wandb_api_context:
            wandb_ctx = self.wandb_api_context.to_json()
        return {
            "trace_context": trace_context,
            "wandb_api_context": wandb_ctx,
        }


# ServerRequest represents a request object sent to the server
@dataclasses.dataclass
class ServerRequest:
    client_id: str
    name: str
    args: typing.Tuple
    context: ServerRequestContext
    id: int = 0

    @classmethod
    def from_json(cls, json: typing.Any) -> "ServerRequest":
        json["context"] = ServerRequestContext.from_json(json["context"])
        return cls(**json)

    def to_json(self) -> typing.Any:
        return {
            "name": self.name,
            "args": self.args,
            "context": self.context.to_json(),
            "id": self.id,
        }

    def error_response(
        self, http_error_code: int, error: Exception
    ) -> "ServerResponse":
        return ServerResponse(
            http_error_code=http_error_code,
            client_id=self.client_id,
            id=self.id,
            # TODO(DG): this is a hack, we should be able to serialize the exception
            value=str(error),
            error=True,
        )

    def success_response(self, value: typing.Any) -> "ServerResponse":
        return ServerResponse(
            http_error_code=200,
            error=False,
            client_id=self.client_id,
            id=self.id,
            value=value,
        )


# ServerResponse represents a response object returned by the server
@dataclasses.dataclass
class ServerResponse:
    client_id: str
    id: int
    value: typing.Any
    error: bool = False
    http_error_code: typing.Optional[int] = None
    http_error_message: typing.Optional[str] = None

    @classmethod
    def from_json(cls, json: typing.Any) -> "ServerResponse":
        return cls(**json)

    def to_json(self) -> typing.Any:
        return {
            "id": self.id,
            "value": self.value,
            "error": self.error,
            "http_error_code": self.http_error_code,
        }


class ShutDown:
    def __eq__(self, other: Any) -> bool:
        return isinstance(other, ShutDown)


shutdown_request = ServerRequest("", "shutdown", (), ServerRequestContext(None, None))
shutdown_response = ServerResponse("", 0, ShutDown())

HandlerFunction = Callable[..., Any]


class HandlerNotFoundException(Exception):
    pass


# Server class is responsible for managing server lifecycle and handling requests
class Server:
    def __init__(
        self,
        process: bool = False,
    ) -> None:
        self.handlers: Dict[str, HandlerFunction] = {}

        self.request_handler: typing.Union[threading.Thread, aioprocessing.AioProcess]
        self.request_queue: async_queue.Queue[ServerRequest]

        # The internal response queue is used to communicate results back between the server
        # process and the user process. The server process puts responses into this queue,
        # and then the Queue feeder thread puts them into the appropriate client response queue.
        self._internal_response_queue: async_queue.Queue[ServerResponse]

        # just using a ThreadQueue here since this is for communication between two threads
        # that are both in the user process.
        self.client_response_queues: Dict[
            str,
            typing.Union[
                queue.Queue[ServerResponse], async_queue.ThreadQueue[ServerResponse]
            ],
        ] = {}

        self._shutting_down = multiprocessing.Event()
        self._shutdown_lock = multiprocessing.Lock()

        self._request_handler_ready_event = multiprocessing.Event()
        self._request_handler_ready_to_shut_down_event = multiprocessing.Event()

        self._response_queue_feeder_ready_event = threading.Event()
        self._response_queue_feeder_ready_to_shut_down_event = threading.Event()

        # Register handlers
        self.register_handler_fn("ensure_manifest", self.handle_ensure_manifest)
        self.register_handler_fn(
            "ensure_file_downloaded", self.handle_ensure_file_downloaded
        )
        self.register_handler_fn("ensure_file", self.handle_ensure_file)
        self.register_handler_fn("direct_url", self.handle_direct_url)
        self.register_handler_fn("sleep", self.handle_sleep)

        if process:
            self.request_handler = aioprocessing.AioProcess(
                target=self._request_handler_fn, name="IO Server", daemon=True
            )
            self.request_queue = async_queue.ProcessQueue()
            self._internal_response_queue = async_queue.ProcessQueue()
        else:
            self.request_handler = threading.Thread(
                target=self._request_handler_fn, name="IO Server", daemon=True
            )
            self.request_queue = async_queue.ThreadQueue()
            self._internal_response_queue = async_queue.ThreadQueue()

        # runs in the user process and puts responses from the server into the appropriate
        # client-consumed response queues.
        self.response_queue_router = threading.Thread(
            target=self._response_queue_router_fn, daemon=True
        )

    # server_process runs the server's main coroutine
    def _request_handler_fn(self) -> None:
        try:
            asyncio.run(self._request_handler_fn_main(), debug=True)
        except Exception as e:
            logging.exception(f"Error in request handler process: {e}")
            raise e

    # start starts the server thread or process
    def start(self) -> None:
        self.request_handler.start()
        self.response_queue_router.start()
        self._request_handler_ready_event.wait()
        self._response_queue_feeder_ready_event.wait()
        atexit.register(self.shutdown)

    # cleanup performs cleanup actions, such as flushing stats
    def cleanup(self) -> None:
        statsd.flush()

    # shutdown stops the server and joins the thread/process
    def shutdown(self) -> None:
        with self._shutdown_lock:
            self._shutting_down.set()

            # tell the two auxiliary processes to shutdown
            # (the server process and the response queue feeder thread)
            self.request_queue.put(shutdown_request)
            self._internal_response_queue.put(shutdown_response)

            self._response_queue_feeder_ready_to_shut_down_event.wait()
            self._request_handler_ready_to_shut_down_event.wait()

            self.response_queue_router.join()
            self.request_handler.join()
            self.cleanup()

    def _response_queue_router_fn(self) -> None:
        try:
            asyncio.run(self._response_queue_router_fn_main(), debug=True)
        except Exception as e:
            logging.exception(f"Error in response queue router: {e}")
            raise e

    async def _response_queue_router_fn_main(self) -> None:
        self._response_queue_feeder_ready_event.set()
        while True:
            try:
                resp = await self._internal_response_queue.async_get()
            except RuntimeError:
                # this happens when the interpreter is shutting down
                break
            if resp.value == ShutDown():
                self._internal_response_queue.task_done()
                self._internal_response_queue.join()
                break
            client_response_queue = self.client_response_queues[resp.client_id]
            # this is non-blocking b/c resp is already in memory
            client_response_queue.put(resp)
            self._internal_response_queue.task_done()
        # drain queue
        self._response_queue_feeder_ready_to_shut_down_event.set()

    async def _handle(self, req: ServerRequest) -> None:
        with tracer.trace("WBArtifactManager.handle.%s" % req.name, service="weave-am"):
            try:
                handler = self.handlers[req.name]
            except KeyError as e:
                resp = req.error_response(404, e)
            else:
                try:
                    val = await handler(*req.args)
                except Exception as e:
                    logging.error(
                        "WBArtifactManager request error: %s\n",
                        traceback.format_exc(),
                    )
                    print(
                        "WBArtifactManager request error: %s\n",
                        traceback.format_exc(),
                    )
                    resp = req.error_response(
                        server_error_handling.maybe_extract_code_from_exception(e)
                        or 500,
                        e,
                    )
                else:
                    resp = req.success_response(val)
            if isinstance(self._internal_response_queue, async_queue.ThreadQueue):
                # this is fast
                self._internal_response_queue.put(resp)
            else:
                # this needs to perform IPC
                await self._internal_response_queue.async_put(resp)

    # main is the server's main coroutine, handling incoming requests
    async def _request_handler_fn_main(self) -> None:
        # This loop should never block.
        fs = filesystem.FilesystemAsync()
        net = weave_http.HttpAsync(fs)
        loop = asyncio.get_running_loop()
        active_tasks: set[asyncio.Task[typing.Any]] = set()
        async with net:
            self.wandb_file_manager = wandb_file_manager.WandbFileManagerAsync(
                fs, net, await wandb_api.get_wandb_api()
            )

            self._request_handler_ready_event.set()
            while True:
                try:
                    req = await self.request_queue.async_get()
                except RuntimeError:
                    # this happens when the loop is shutting down
                    break
                if req.name == "shutdown":
                    self.request_queue.task_done()
                    self.request_queue.join()
                    break
                tracer.context_provider.activate(req.context.trace_context)
                with wandb_api.wandb_api_context(req.context.wandb_api_context):
                    # launch a task to handle the request
                    task = loop.create_task(self._handle(req))
                    active_tasks.add(task)
                    task.add_done_callback(active_tasks.discard)
                self.request_queue.task_done()
        self._request_handler_ready_to_shut_down_event.set()

    def register_handler_fn(self, name: str, handler: HandlerFunction) -> None:
        self.handlers[name] = handler

    @contextlib.contextmanager
    def registered_client(
        self, client: typing.Union["AsyncClient", "SyncClient"]
    ) -> Iterator[None]:
        self.register_client(client)
        try:
            yield
        finally:
            self.unregister_client(client)

    def register_client(
        self, client: typing.Union["SyncClient", "AsyncClient"]
    ) -> None:
        if client.client_id not in self.client_response_queues:
            if isinstance(client, SyncClient):
                self.client_response_queues[client.client_id] = queue.Queue()
            else:
                self.client_response_queues[
                    client.client_id
                ] = async_queue.ThreadQueue()

    def unregister_client(
        self, client: typing.Union["SyncClient", "AsyncClient"]
    ) -> None:
        if client.client_id in self.client_response_queues:
            del self.client_response_queues[client.client_id]

    async def handle_ensure_manifest(
        self, artifact_uri: str
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        uri = artifact_wandb.WeaveWBArtifactURI.parse(artifact_uri)
        return await self.wandb_file_manager.manifest(uri)

    async def handle_ensure_file(self, artifact_uri: str) -> typing.Optional[str]:
        uri = artifact_wandb.WeaveWBArtifactURI.parse(artifact_uri)
        return await self.wandb_file_manager.ensure_file(uri)

    async def handle_ensure_file_downloaded(
        self, download_url: str
    ) -> typing.Optional[str]:
        return await self.wandb_file_manager.ensure_file_downloaded(download_url)

    async def handle_direct_url(self, artifact_uri: str) -> typing.Optional[str]:
        uri = artifact_wandb.WeaveWBArtifactURI.parse(artifact_uri)
        return await self.wandb_file_manager.direct_url(uri)

    async def handle_sleep(self, seconds: float) -> float:
        # used for testing to simulate long running processes
        await asyncio.sleep(seconds)
        return seconds


SERVER = None
SERVER_START_LOCK = threading.Lock()


def get_server() -> Server:
    global SERVER
    with SERVER_START_LOCK:
        if SERVER is None:
            SERVER = Server(process=False)
            SERVER.start()
        return SERVER


class AsyncConnection:
    def __init__(
        self,
        client_id: str,
        server: Server,
    ) -> None:
        self.client_id = client_id
        self.server = server
        response_queue = server.client_response_queues[client_id]
        request_queue = server.request_queue

        self.request_id = 0
        self.requests: typing.Dict[int, asyncio.Future] = {}
        self.request_queue: async_queue.Queue[ServerRequest] = request_queue
        self.response_queue: async_queue.ThreadQueue[ServerResponse] = typing.cast(
            async_queue.ThreadQueue[ServerResponse], response_queue
        )
        self.response_task = asyncio.create_task(self.handle_responses())
        self.response_task.add_done_callback(self.response_task_ended)
        self.connected = True

    def response_task_ended(self, task: asyncio.Task) -> None:
        exc = task.exception()
        if exc:
            print("IOServiceError", exc)
            traceback.print_exception(type(exc), exc, exc.__traceback__)
            raise exc

    async def close(self) -> None:
        self.connected = False

    async def handle_responses(self) -> None:
        while self.connected:
            resp = await self.response_queue.async_get()
            self.response_queue.task_done()
            self.requests[resp.id].set_result(resp)

    async def request(self, name: str, *args: typing.Any) -> typing.Any:
        # Caller must check ServerResponse.error!

        wb_ctx = wandb_api.get_wandb_api_context()
        cur_trace_context = tracer.current_trace_context()

        req = ServerRequest(
            self.client_id,
            name,
            args,
            ServerRequestContext(cur_trace_context, wb_ctx),
            self.request_id,
        )

        self.request_id += 1
        response_future: asyncio.Future[ServerResponse] = asyncio.Future()
        self.requests[req.id] = response_future

        with self.server._shutdown_lock:
            is_shutting_down = self.server._shutting_down.is_set()

        if is_shutting_down:
            self.response_task.cancel()
            raise errors.WeaveWandbArtifactManagerError(
                "Server is shutting down, cannot make request"
            )

        if isinstance(self.request_queue, async_queue.ThreadQueue):
            # this is fast
            self.request_queue.put(req)
        else:
            await self.request_queue.async_put(req)
        server_resp = await response_future

        if server_resp.error:
            if server_resp.http_error_code != None:
                raise server_error_handling.WeaveInternalHttpException.from_code(
                    server_resp.http_error_code
                )
            raise errors.WeaveWandbArtifactManagerError(
                "Request error: " + server_resp.value
            )

        return server_resp.value

    async def manifest(
        self, artifact_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        manifest: typing.Optional[
            artifact_wandb.WandbArtifactManifest
        ] = await self.request("ensure_manifest", str(artifact_uri))
        return manifest

    async def ensure_file(
        self, artifact_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[str]:
        res = await self.request("ensure_file", str(artifact_uri))
        return res

    async def ensure_file_downloaded(self, download_url: str) -> typing.Optional[str]:
        res = await self.request("ensure_file_downloaded", download_url)
        return res

    async def direct_url(
        self, artifact_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[str]:
        res = await self.request("direct_url", str(artifact_uri))
        return res

    async def sleep(self, seconds: float) -> float:
        return await self.request("sleep", seconds)


class AsyncClient:
    def __init__(self, server: Server) -> None:
        self.client_id = str(uuid.uuid4())
        self.server = server

    @contextlib.asynccontextmanager
    async def connect(self) -> typing.AsyncGenerator[AsyncConnection, None]:
        with self.server.registered_client(self):
            conn = AsyncConnection(self.client_id, self.server)
            try:
                yield conn
            finally:
                await conn.close()


class SyncClient:
    def __init__(self, server: Server, fs: filesystem.Filesystem) -> None:
        self.client_id = str(uuid.uuid4())
        self.fs = fs
        self.server = server
        self._current_request_id = 0

    def request(self, name: str, *args: typing.Any) -> typing.Any:
        wb_ctx = wandb_api.get_wandb_api_context()
        cur_trace_context = tracer.current_trace_context()
        self._current_request_id += 1

        with self.server.registered_client(self):
            request = ServerRequest(
                self.client_id,
                name,
                args,
                ServerRequestContext(cur_trace_context, wb_ctx),
                id=self._current_request_id,
            )

            response_queue = self.server.client_response_queues[self.client_id]

            with self.server._shutdown_lock:
                is_shutting_down = self.server._shutting_down.is_set()

            if is_shutting_down:
                raise errors.WeaveWandbArtifactManagerError(
                    "Server is shutting down, cannot make request"
                )

            self.server.request_queue.put(request)
            server_resp = response_queue.get()
            response_queue.task_done()

        if server_resp.error:
            if server_resp.http_error_code != None:
                raise server_error_handling.WeaveInternalHttpException.from_code(
                    server_resp.http_error_code
                )
            raise errors.WeaveWandbArtifactManagerError(
                "Request error: " + server_resp.value
            )
        return server_resp.value

    def manifest(
        self, artifact_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        manifest: typing.Optional[artifact_wandb.WandbArtifactManifest] = self.request(
            "ensure_manifest", str(artifact_uri)
        )
        return manifest

    def ensure_file(
        self, artifact_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[str]:
        return self.request("ensure_file", str(artifact_uri))

    def ensure_file_downloaded(self, download_url: str) -> typing.Optional[str]:
        return self.request("ensure_file_downloaded", download_url)

    def direct_url(
        self, artifact_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[str]:
        return self.request("direct_url", str(artifact_uri))

    def sleep(self, seconds: float) -> None:
        return self.request("sleep", seconds)


class ServerlessClient:
    def __init__(self, fs: filesystem.Filesystem) -> None:
        self.fs = fs
        self.http = weave_http.Http(self.fs)
        self.wandb_api = wandb_api.WandbApi()
        self.wandb_file_manager = wandb_file_manager.WandbFileManager(
            self.fs, self.http, self.wandb_api
        )

    def manifest(
        self, artifact_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        return self.wandb_file_manager.manifest(artifact_uri)

    def ensure_file(
        self, artifact_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[str]:
        return self.wandb_file_manager.ensure_file(artifact_uri)

    def ensure_file_downloaded(self, download_url: str) -> typing.Optional[str]:
        return self.wandb_file_manager.ensure_file_downloaded(download_url)

    def direct_url(
        self, artifact_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[str]:
        return self.wandb_file_manager.direct_url(artifact_uri)

    def sleep(self, seconds: float) -> None:
        time.sleep(seconds)


def get_sync_client() -> typing.Union[SyncClient, ServerlessClient]:
    # uncomment this for local debugging
    # return ServerlessClient(filesystem.get_filesystem())
    return SyncClient(get_server(), filesystem.get_filesystem())


def get_async_client() -> AsyncClient:
    return AsyncClient(get_server())



from weave.server import handle_request


def test_playback():
    from weave import ecosystem

    for payload in [execute_payloads[-1]]:
        res = handle_request(payload, True)
        res.results.unwrap()


execute_payloads = [
    {
        "graphs": {
            "nodes": [
                {"nodeType": "const", "type": "string", "val": "inherit"},
                {"nodeType": "const", "type": "string", "val": ""},
                {
                    "nodeType": "output",
                    "fromOp": 3,
                    "type": "number",
                    "id": "6965691728254189",
                },
                {"name": "tag-indexCheckpoint", "inputs": {"obj": 4}},
                {
                    "nodeType": "output",
                    "fromOp": 5,
                    "type": {
                        "type": "tagged",
                        "tag": {
                            "type": "typedDict",
                            "propertyTypes": {"indexCheckpoint": "number"},
                        },
                        "value": {"type": "OpDef"},
                    },
                    "id": "5817702282560567",
                },
                {"name": "index", "inputs": {"arr": 6, "index": 12}},
                {
                    "nodeType": "output",
                    "fromOp": 7,
                    "type": {
                        "type": "list",
                        "objectType": {
                            "type": "tagged",
                            "tag": {
                                "type": "typedDict",
                                "propertyTypes": {"indexCheckpoint": "number"},
                            },
                            "value": {"type": "OpDef"},
                        },
                    },
                    "id": "2857565887621620",
                },
                {"name": "list-createIndexCheckpointTag", "inputs": {"arr": 8}},
                {
                    "nodeType": "output",
                    "fromOp": 9,
                    "type": {"type": "list", "objectType": {"type": "OpDef"}},
                    "id": "3688357283448756",
                },
                {"name": "Ecosystem-ops", "inputs": {"self": 10}},
                {
                    "nodeType": "output",
                    "fromOp": 11,
                    "type": {
                        "type": "Ecosystem",
                        "_orgs": {"type": "list", "objectType": "any"},
                        "_packages": {"type": "list", "objectType": {"type": "OpDef"}},
                        "_datasets": {"type": "list", "objectType": {"type": "OpDef"}},
                        "_models": {"type": "list", "objectType": "any"},
                        "_ops": {"type": "list", "objectType": {"type": "OpDef"}},
                        "_property_types": {
                            "_orgs": {"type": "list", "objectType": "any"},
                            "_packages": {
                                "type": "list",
                                "objectType": {"type": "OpDef"},
                            },
                            "_datasets": {
                                "type": "list",
                                "objectType": {"type": "OpDef"},
                            },
                            "_models": {"type": "list", "objectType": "any"},
                            "_ops": {"type": "list", "objectType": {"type": "OpDef"}},
                        },
                    },
                    "id": "8371329692561700",
                },
                {"name": "op-ecosystem", "inputs": {}},
                {"nodeType": "const", "type": "number", "val": 0},
                {
                    "nodeType": "output",
                    "fromOp": 14,
                    "type": "number",
                    "id": "5907007908746573",
                },
                {"name": "tag-indexCheckpoint", "inputs": {"obj": 15}},
                {
                    "nodeType": "output",
                    "fromOp": 16,
                    "type": {
                        "type": "tagged",
                        "tag": {
                            "type": "typedDict",
                            "propertyTypes": {"indexCheckpoint": "number"},
                        },
                        "value": {"type": "OpDef"},
                    },
                    "id": "6457218123539456",
                },
                {"name": "index", "inputs": {"arr": 6, "index": 17}},
                {"nodeType": "const", "type": "number", "val": 1},
                {
                    "nodeType": "output",
                    "fromOp": 19,
                    "type": "number",
                    "id": "6997650288408352",
                },
                {"name": "tag-indexCheckpoint", "inputs": {"obj": 20}},
                {
                    "nodeType": "output",
                    "fromOp": 21,
                    "type": {
                        "type": "tagged",
                        "tag": {
                            "type": "typedDict",
                            "propertyTypes": {"indexCheckpoint": "number"},
                        },
                        "value": {"type": "OpDef"},
                    },
                    "id": "3586667514912185",
                },
                {"name": "index", "inputs": {"arr": 6, "index": 22}},
                {"nodeType": "const", "type": "number", "val": 2},
                {
                    "nodeType": "output",
                    "fromOp": 24,
                    "type": "number",
                    "id": "7805106646043236",
                },
                {"name": "tag-indexCheckpoint", "inputs": {"obj": 25}},
                {
                    "nodeType": "output",
                    "fromOp": 26,
                    "type": {
                        "type": "tagged",
                        "tag": {
                            "type": "typedDict",
                            "propertyTypes": {"indexCheckpoint": "number"},
                        },
                        "value": {"type": "OpDef"},
                    },
                    "id": "2587861762122901",
                },
                {"name": "index", "inputs": {"arr": 6, "index": 27}},
                {"nodeType": "const", "type": "number", "val": 3},
                {
                    "nodeType": "output",
                    "fromOp": 29,
                    "type": "number",
                    "id": "5057421628670159",
                },
                {"name": "count", "inputs": {"arr": 6}},
            ],
            "targetNodes": [0, 1, 2, 13, 18, 23, 28],
        }
    }
]



# Trace Weave engine with Datadog
#
# There are four trace modes:
#   DD_ENV & !WEAVE_TRACE_STREAM: log to datadog only
#   [experimental] DD_ENV & WEAVE_TRACE_STREAM: mirror datadog spans to a W&B StreamTable
#     this is good because it uses all the datadog instrumentation
#   [experimental] !DD_ENV & WEAVE_TRACE_STREAM: log trace tree traces to a W&B StreamTable
#     only captures trace calls, not other library instrumentation as added by datadog
#   None: DummyTrace that does nothing


import os
import typing
import contextvars
import multiprocessing
import logging
import time
import json
import dataclasses

from . import logs


# Thanks co-pilot!
class DummySpan:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.log_indent_token = None

    def __enter__(self):
        logging.debug("-> %s", self.args[0])
        self.log_indent_token = logs.increment_indent()
        return self

    def __exit__(self, *args, **kwargs):
        logs.reset_indent(self.log_indent_token)
        logging.debug("<- %s", self.args[0])

    def set_tag(self, *args, **kwargs):
        pass

    def set_meta(self, *args, **kwargs):
        pass

    def finish(self, *args, **kwargs):
        pass


class TraceContext:
    def __getstate__(self) -> dict:
        return {}

    def __setstate__(self, state: dict) -> None:
        pass


class ContextProvider:
    def activate(self, context: TraceContext) -> None:
        pass


class DummyTrace:
    def trace(self, *args, **kwargs):
        return DummySpan(*args)

    @property
    def context_provider(self) -> ContextProvider:
        return ContextProvider()

    def current_trace_context(self) -> typing.Optional[TraceContext]:
        return None

    def current_span(self):
        return None

    def current_root_span(self):
        return None


_weave_trace_span: contextvars.ContextVar[
    typing.Optional["WeaveTraceSpan"]
] = contextvars.ContextVar("_weave_trace_span", default=None)


_weave_trace_stream = None


def weave_trace_stream():
    global _weave_trace_stream
    if _weave_trace_stream is None:
        from weave.wandb_interface.wandb_stream_table import StreamTable

        _weave_trace_stream = StreamTable(os.getenv("WEAVE_TRACE_STREAM"))
    return _weave_trace_stream


def span_count(span):
    if span.child_spans is None:
        return 1
    return 1 + sum([span_count(child) for child in span.child_spans])


class WeaveTraceSpan:
    def __init__(self, name):
        from .ops_domain import trace_tree

        self.log_indent_token = None
        self._token = None
        self.span = trace_tree.Span(_name=name)
        self._parent = None

    def total_span_count(self):
        return span_count(self.span)

    def __enter__(self):
        parent = _weave_trace_span.get()
        if parent is not None:
            if parent.span.child_spans is None:
                parent.span.child_spans = []
            parent.span.child_spans.append(self.span)
            self._parent = parent

        self._token = _weave_trace_span.set(self)
        self.span.start_time_ms = int(time.time() * 1000)
        return self

    def __exit__(self, *args, **kwargs):
        self.span.end_time_ms = int(time.time() * 1000)
        _weave_trace_span.reset(self._token)

        from .ops_domain import trace_tree

        tt = trace_tree.WBTraceTree(json.dumps(dataclasses.asdict(self.span)))
        tags = self.attributes.get("tags", {})
        if _weave_trace_span.get() is None:
            weave_trace_stream().log(
                {
                    "span_count": self.total_span_count(),
                    "duration_ms": self.span.end_time_ms - self.span.start_time_ms,
                    **tags,
                    "trace": tt,
                }
            )

    @property
    def attributes(self):
        if self.span.attributes is None:
            self.span.attributes = {}
        return self.span.attributes

    def set_tag(self, key, val):
        if "tags" not in self.attributes:
            self.attributes["tags"] = {}
        self.attributes["tags"][key] = val

    def set_meta(self, key, val):
        if "metadata" not in self.attributes:
            self.attributes["metadata"] = {}
        self.attributes["metadata"][key] = val

    def finish(self, *args, **kwargs):
        pass


class WeaveTrace:
    def __init__(self):
        self._stream_table = None

    def trace(self, name, *args, **kwargs):
        return WeaveTraceSpan(name, *args)

    @property
    def context_provider(self) -> ContextProvider:
        return ContextProvider()

    def current_trace_context(self) -> typing.Optional[TraceContext]:
        return None

    def current_span(self):
        return _weave_trace_span.get()

    def current_root_span(self):
        cur_span = _weave_trace_span.get()
        while cur_span is not None and cur_span._parent is not None:
            cur_span = cur_span._parent
        return cur_span


def dd_span_to_weave_span(dd_span):
    # Use '' for None, currently history2 doesn't read None columns from
    # the liveset correctly.
    parent_id = ""
    if dd_span.parent_id is not None:
        parent_id = str(dd_span.parent_id)
    return {
        "name": dd_span.name,
        "start_time_ms": dd_span.start_ns / 1e6,
        "end_time_ms": (dd_span.start_ns + dd_span.duration_ns) / 1e6,
        "attributes": dd_span.get_tags(),
        "trace_id": str(dd_span.trace_id),
        "span_id": str(dd_span.span_id),
        "parent_id": parent_id,
    }


def send_proc(queue):
    while True:
        spans = queue.get()
        if spans is None:
            break
        trace_stream = weave_trace_stream()
        if trace_stream is not None:
            for span in spans:
                trace_stream.log(span)


# A DataDog writer that sends spans to W&B as a StreamTable.
# We have to use a separate process for writing the stream, otherwise things hang.
# My guess is this is because logging to StreamTable uses gql, which is wrapped
# by datadog, so we have some kind of re-entrancy/deadlock. Putting in a separate
# process fixes.
class WeaveWriter:
    def __init__(self, orig_writer):
        self._orig_writer = orig_writer
        self._queue = multiprocessing.Queue()
        self._proc = multiprocessing.Process(target=send_proc, args=(self._queue,))

    def recreate(self):
        return WeaveWriter(self._orig_writer.recreate())

    def stop(self, timeout=None):
        self._orig_writer.stop(timeout)

    def _ensure_started(self):
        if not self._proc.is_alive():
            self._proc.start()

    def write(self, spans):
        if len(spans) == 1:
            return
        self._ensure_started()
        self._queue.put([dd_span_to_weave_span(s) for s in spans])
        self._orig_writer.write(spans)

    def flush_queue(self):
        self._orig_writer.flush_queue()


def tracer():

    if os.getenv("DD_ENV"):
        from ddtrace import tracer as ddtrace_tracer

        if os.getenv("WEAVE_TRACE_STREAM"):
            # In DataDog mode, if WEAVE_TRACE_STREAM is set, experimentally
            # mirror DataDog trace info to W&B.
            # In this mode we log a table of spans, as opposed to traces.
            if not isinstance(ddtrace_tracer._writer, WeaveWriter):
                ddtrace_tracer._writer = WeaveWriter(ddtrace_tracer._writer)
                from ddtrace.internal.processor.trace import SpanAggregator

                span_agg = ddtrace_tracer._span_processors[-1]
                if not isinstance(span_agg, SpanAggregator):
                    raise ValueError("unable to patch ddtrace tracer")
                span_agg._writer = ddtrace_tracer._writer

        return ddtrace_tracer
    elif os.getenv("WEAVE_TRACE_STREAM"):
        # Without DataDog, WEAVE_TRACE_STREAM tells us to log any trace calls as a trace tree
        return WeaveTrace()
    else:
        return DummyTrace()


def new_trace_context() -> typing.Optional[TraceContext]:
    if os.getenv("DD_ENV"):
        from ddtrace import context as ddtrace_context

        return ddtrace_context.Context()  # type: ignore
    else:
        return None


class DummyStatsd:
    def __init__(self, *args, **kwargs):
        pass

    def increment(self, *args, **kwargs):
        pass

    def decrement(self, *args, **kwargs):
        pass

    def gauge(self, *args, **kwargs):
        pass

    def flush(self, *args, **kwargs):
        pass


_STATSD = None


def _initialize_statsd():
    if os.getenv("DD_ENV"):
        from datadog import initialize, statsd

        initialize(statsd_disable_buffering=False)
        return statsd
    else:
        return DummyStatsd()


def statsd():
    global _STATSD
    if _STATSD is None:
        _STATSD = _initialize_statsd()

    return _STATSD


def datadog_is_enabled():
    return os.getenv("DD_ENV")



from . import gpt3

from ... import weave_types as types
from .. import openai
from ... import api as weave
from ... import storage
from ... import graph
from .. import panels
from ...show import _show_params


def test_gpt3model_inferred_type():
    assert gpt3.Gpt3Model.complete.input_type.arg_types == {
        "self": gpt3.Gpt3ModelType(),
        "prompt": types.String(),
    }
    assert gpt3.Gpt3Model.complete.concrete_output_type == types.TypedDict(
        {
            "id": types.String(),
            "object": types.String(),
            "created": types.Int(),
            "model": types.String(),
            "choices": types.List(
                types.TypedDict(
                    {
                        "text": types.String(),
                        "index": types.Int(),
                        "logprobs": types.none_type,
                        "finish_reason": types.String(),
                    }
                )
            ),
        }
    )


def test_large_const_node(test_artifact_dir):
    data = []
    for i in range(500):
        a = i
        b = i % 9
        r = a + b
        data.append(
            {"id": i, "prompt": "%s + %s =" % (a, b), "completion": " %s<end>" % r}
        )
    dataset = storage.save(data)
    # storage.save(dataset)
    fine_tune = openai.finetune_gpt3(dataset, {"n_epochs": 2})
    actual = _show_params(fine_tune)["weave_node"].to_json()

    print("test_large_const_node.actual", actual)

    ACTUAL_SHOW_PARAMS_FINE_TUNE_WEAVE_NODE = {
        "nodeType": "output",
        "type": "any",
        "fromOp": {
            "name": "get",
            "inputs": {
                "uri": {
                    "nodeType": "const",
                    "type": "string",
                    "val": "local-artifact:///dashboard-Run:latest/obj",
                }
            },
        },
    }
    assert actual == ACTUAL_SHOW_PARAMS_FINE_TUNE_WEAVE_NODE

    model = openai.Gpt3FineTune.model(fine_tune)
    panel = panels.Table(["1 + 9 =", "2 + 14 ="])
    panel.config.tableState.add_column(lambda row: row)
    panel.config.tableState.add_column(
        lambda row: model.complete(row)["choices"][0]["text"]
    )
    show_panel_params = _show_params(panel)
    panel_params = weave.use(show_panel_params["weave_node"])

    panel_config = panel_params.config
    # Ensure that we sent the dataset as a get(<ref>) rather than as a const list
    # (this behavior is currently implemented in graph.py:ConstNode)

    table_state = panel_config.tableState
    col_select_fns = table_state.columnSelectFunctions
    col_sel_fn2 = list(col_select_fns.values())[1]
    assert "list:" in graph.node_expr_str(col_sel_fn2)

    # Asserting that weavejs_fixes.remove_opcall_versions_data works
    assert (
        graph.node_expr_str(col_sel_fn2)
        == 'finetune_gpt3(local-artifact:///list:f909a3e7a090a0a57dbd/obj, {"n_epochs": 2}).model().complete(row)["choices"][0]["text"]'
    )



import typing
import contextvars
import contextlib
import inspect
import typing_extensions
import dataclasses
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import textwrap

from .. import ref_base
from .. import weave_types as types
from .. import box
from .. import weave_internal
from .. import errors
from .. import graph
from ..language_features.tagging import (
    tagged_value_type,
)
from .. import artifact_base
from .. import node_ref
from ..language_features.tagging import tag_store

from .arrow import (
    safe_is_null,
    ArrowWeaveListType,
    offsets_starting_at_zero,
    pretty_print_arrow_type,
    arrow_zip,
    arrow_as_array,
)
from .. import debug_types


def reverse_dict(d: dict) -> dict:
    return dict(reversed(d.items()))


def concrete_to_tagstore(val: typing.Any) -> typing.Any:
    if isinstance(val, dict):
        if "_tag" in val and "_value" in val:
            v = box.box(concrete_to_tagstore(val["_value"]))
            tag_store.add_tags(v, concrete_to_tagstore(val["_tag"]))
            return v
        return {k: concrete_to_tagstore(v) for k, v in val.items()}
    elif isinstance(val, list):
        return [concrete_to_tagstore(v) for v in val]
    elif dataclasses.is_dataclass(val):
        params = {
            f.name: concrete_to_tagstore(getattr(val, f.name))
            for f in dataclasses.fields(val)
        }
        return val.__class__(**params)
    return val


def diff_arrow_type(t1, t2):
    if t1 == t2:
        return None
    if (
        isinstance(t1, pa.StructType)
        and isinstance(t2, pa.StructType)
        or isinstance(t1, pa.UnionType)
        and isinstance(t2, pa.UnionType)
    ):
        if isinstance(t1, pa.StructType):
            error_prefix = "structs"
        else:
            error_prefix = "unions"
        if t1.num_fields != t2.num_fields:
            return f"{error_prefix} different number of fields"
        for f1, f2 in zip(t1, t2):
            if f1.name != f2.name:
                return f"{error_prefix} different field names: {f1.name} != {f2.name}"
            diff = diff_arrow_type(f1.type, f2.type)
            if diff is not None:
                return f"{error_prefix} field {f1.name}: {diff}"
        return None
    if isinstance(t1, pa.ListType) and isinstance(t2, pa.ListType):
        return diff_arrow_type(t1.value_type, t2.value_type)
    return f"{t1}\n!=\n{t2}"


def first_non_none(v):
    for i in v:
        if i is not None:
            return i
    return None


@dataclasses.dataclass(frozen=True)
class PathItemOuterList:
    # Denotes this is the outermost list
    pass


@dataclasses.dataclass(frozen=True)
class PathItemStructField:
    key: str


@dataclasses.dataclass(frozen=True)
class PathItemObjectField:
    attr: str


@dataclasses.dataclass(frozen=True)
class PathItemList:
    pass


@dataclasses.dataclass(frozen=True)
class PathItemTaggedValueTag:
    pass


@dataclasses.dataclass(frozen=True)
class PathItemTaggedValueValue:
    pass


@dataclasses.dataclass(frozen=True)
class PathItemUnionEntry:
    type_code: int


PathItemType = typing.Union[
    PathItemOuterList,
    PathItemStructField,
    PathItemObjectField,
    PathItemList,
    PathItemTaggedValueTag,
    PathItemTaggedValueValue,
    PathItemUnionEntry,
]

PathType = tuple[PathItemType, ...]


def set_path(
    v: typing.Any,
    arrow_data: pa.Array,
    path: PathType,
    value_fn: typing.Callable[[typing.Any, int], typing.Any],
    offset: int = 0,
    arrow_offset: int = 0,
):
    last_settable_index = len(path) - 1
    final_union_path_el: typing.Optional[PathItemUnionEntry] = None
    if isinstance(path[last_settable_index], PathItemUnionEntry):
        last_settable_index -= 1
        final_union_path_el = path[-1]  # type: ignore
    if isinstance(path[last_settable_index], PathItemUnionEntry):
        raise errors.WeaveInternalError("path ended with two unions, invalid")
    # Helper function for converting arrow to python. Set the value at a path
    # within the value v, using value_fn to get the new value.
    # Path must be of length 1 or greater. Caller needs to handle the case
    # where we want to set v itself (path is of length 0).
    for i, el in enumerate(path[:last_settable_index]):
        if v is None:
            return
        if isinstance(el, PathItemStructField):
            v = v[el.key]
            arrow_data = arrow_data.field(el.key)
        elif isinstance(el, PathItemObjectField):
            v = getattr(v, el.attr)
            arrow_data = arrow_data.field(el.attr)
        elif el == PathItemTaggedValueTag():
            try:
                v = v["_tag"]
            except (TypeError, KeyError):
                v = v.tag
            arrow_data = arrow_data.field("_tag")
        elif el == PathItemTaggedValueValue():
            try:
                v = v["_value"]
            except (TypeError, KeyError):
                v = v.value
            arrow_data = arrow_data.field("_value")
        elif el == PathItemOuterList():
            for j, vi in enumerate(v):
                set_path(
                    vi,
                    arrow_data,
                    path[i + 1 :],
                    value_fn,
                    offset + j,
                    arrow_offset + j,
                )
            return
        elif el == PathItemList():
            # its a list
            for j, vi in enumerate(v):
                set_path(
                    vi,
                    arrow_data.values,
                    path[i + 1 :],
                    value_fn,
                    offset + j,
                    arrow_offset + j,
                )
            return
        elif isinstance(el, PathItemUnionEntry):
            if el.type_code != arrow_data.type_codes[arrow_offset].as_py():
                return
            arrow_data = arrow_data.field(el.type_code)
            arrow_offset = 0
        else:
            raise ValueError(f"Unexpected path element: {el}")
    if v is None:
        return

    leaf = path[last_settable_index]

    if isinstance(leaf, PathItemStructField):
        if final_union_path_el:
            arrow_data = arrow_data.field(leaf.key)
            if (
                final_union_path_el.type_code
                != arrow_data.type_codes[arrow_offset].as_py()
            ):
                return
        v[leaf.key] = value_fn(v[leaf.key], offset)
    elif isinstance(leaf, PathItemObjectField):
        if final_union_path_el:
            arrow_data = arrow_data.field(leaf.attr)
            if (
                final_union_path_el.type_code
                != arrow_data.type_codes[arrow_offset].as_py()
            ):
                return
        setattr(v, leaf.attr, value_fn(getattr(v, leaf.attr), offset))
    elif leaf == PathItemTaggedValueTag():
        if final_union_path_el:
            arrow_data = arrow_data.field("_tag")
            if (
                final_union_path_el.type_code
                != arrow_data.type_codes[arrow_offset].as_py()
            ):
                return
        try:
            v["_tag"] = value_fn(v["_tag"], offset)
        except (KeyError, TypeError):
            v.tag = value_fn(v.tag, offset)
    elif leaf == PathItemTaggedValueValue():
        if final_union_path_el:
            arrow_data = arrow_data.field("_value")
            if (
                final_union_path_el.type_code
                != arrow_data.type_codes[arrow_offset].as_py()
            ):
                return
        try:
            v["_value"] = value_fn(v["_value"], offset)
        except (KeyError, TypeError):
            # print("V", v)
            v.value = value_fn(v.value, offset)
    elif leaf == PathItemList():
        if final_union_path_el:
            arrow_data = arrow_data.values
            for j, vi in enumerate(v):
                if (
                    final_union_path_el.type_code
                    != arrow_data.type_codes[arrow_offset + j].as_py()
                ):
                    continue
                v[j] = value_fn(vi, offset + j)
        else:
            for j, vi in enumerate(v):
                v[j] = value_fn(vi, offset + j)
    elif leaf == PathItemOuterList():
        if final_union_path_el:
            for j, vi in enumerate(v):
                if (
                    final_union_path_el.type_code
                    != arrow_data.type_codes[arrow_offset + j].as_py()
                ):
                    continue
                v[j] = value_fn(vi, offset + j)
        else:
            for j, vi in enumerate(v):
                v[j] = value_fn(vi, offset + j)
    else:
        raise ValueError(f"Unexpected path element: {leaf}")


def weave_arrow_type_check(
    wt: types.Type, arr: pa.Array, optional: bool = False
) -> typing.Optional[str]:
    reasons: list[str] = []
    at: pa.DataType = arr.type
    if wt == types.Any():
        reasons.append("Any not allowed")
    elif isinstance(wt, types.UnionType):
        for i in range(len(wt.members)):
            for j in range(i + 1, len(wt.members)):
                if not isinstance(
                    types.merge_types(wt.members[i], wt.members[j]), types.UnionType
                ):
                    reasons.append(f"Union members {i} and {j} are mergable")
        non_none_members = [m for m in wt.members if m != types.NoneType()]
        none_member_count = len(wt.members) - len(non_none_members)
        if none_member_count > 1:
            reasons.append("Multiple None members in Union")
        elif none_member_count == 1:
            # Nullable case
            # TODO: Actually must check nullable!
            if len(non_none_members) == 0:
                reasons.append("NoneType not allowed in Union with no other members")
            elif len(non_none_members) == 1:
                reason = weave_arrow_type_check(non_none_members[0], arr, optional=True)
            else:
                reason = weave_arrow_type_check(
                    types.UnionType(*non_none_members), arr, optional=True
                )
            if reason is not None:
                reasons.append(f"Nullable member: {reason}")
        else:
            if not pa.types.is_union(at):
                reasons.append(f"Expected UnionType, got {at}")
            else:
                if len(wt.members) != len(at):
                    reasons.append(
                        f"Union length mismatch: {len(wt.members)} != {len(at)}"
                    )
                else:
                    for i, w in enumerate(wt.members):
                        reason = weave_arrow_type_check(
                            w, arr.field(i), optional=optional
                        )
                        if reason is not None:
                            reasons.append(f"Union member {i}: {reason}")
    elif optional and pa.types.is_null(at):
        pass
    else:
        # Nullable check. This surfaces a ton of failures...
        # if wt != types.NoneType() and not optional and arr.null_count > 0:
        #     reasons.append("Non-nullable but array has nulls")

        if isinstance(wt, tagged_value_type.TaggedValueType):
            tag_reason = weave_arrow_type_check(
                wt.tag, arr.field("_tag"), optional=optional
            )
            if tag_reason is not None:
                reasons.append(f"TaggedValue tag: {tag_reason}")
            value_reason = weave_arrow_type_check(
                wt.value, arr.field("_value"), optional=optional
            )
            if value_reason is not None:
                reasons.append(f"TaggedValue value: {value_reason}")

        elif isinstance(wt, types.Const):
            reasons.append("Const not allowed")

        elif isinstance(wt, types.Function):
            if not pa.types.is_string(at):
                reasons.append(f"Expected StringType, got {at}")

        elif isinstance(wt, types.TypedDict):
            if not pa.types.is_struct(at):
                reasons.append(f"Expected StructType, got {at}")
            else:
                at_fields_by_name = {f.name: f for f in at}
                for k, v in wt.property_types.items():
                    if k not in at_fields_by_name:
                        reasons.append(f"Missing field {k}")
                    else:
                        reason = weave_arrow_type_check(
                            v, arr.field(k), optional=optional
                        )
                        if reason is not None:
                            reasons.append(f"Field {k}: {reason}")

        elif isinstance(wt, (types.List, ArrowWeaveListType)):
            if not pa.types.is_list(at):
                reasons.append(f"Expected ListType, got {at}")
            else:
                reason = weave_arrow_type_check(
                    wt.object_type, arr.flatten(), optional=optional
                )
                if reason is not None:
                    reasons.append(f"List element: {reason}")

        elif isinstance(wt, types.ObjectType):
            if not pa.types.is_struct(at):
                reasons.append(f"Expected StructType, got {at}")
            else:
                at_fields_by_name = {f.name: f for f in at}
                for k, v in wt.property_types().items():
                    if k not in at_fields_by_name:
                        reasons.append(f"Missing field {k}")
                    else:
                        reason = weave_arrow_type_check(
                            v, arr.field(k), optional=optional
                        )
                        if reason is not None:
                            reasons.append(f"Field {k}: {reason}")
        elif isinstance(wt, types.String):
            if pa.types.is_dictionary(at):
                if not pa.types.is_string(at.value_type):
                    reasons.append(
                        f"Expected StringType as dictionary type, got {at.value_type}"
                    )
            elif not pa.types.is_string(at):
                reasons.append(f"Expected StringType, got {at}")
        elif isinstance(wt, types.UnknownType):
            if len(arr) != 0:
                reasons.append(f"Non-zero length array with UnknownType")
        elif isinstance(wt, types.Int):
            # We allow float32/64 because of a bug in our unnest implementation
            # which relies on pandas. TODO: Remove this when we fix unnest.
            if (
                not pa.types.is_float64(at)
                and not pa.types.is_int64(at)
                and not pa.types.is_float32(at)
                and not pa.types.is_int32(at)
            ):
                reasons.append(
                    f"Expected int64 or int32 (float32/64 allowed because of unnest bug), got {at}"
                )
        elif isinstance(wt, types.Float):
            if not pa.types.is_float64(at) and not pa.types.is_float32(at):
                reasons.append(f"Expected float64 or float32, got {at}")
        elif isinstance(wt, types.Number):
            if (
                not pa.types.is_float64(at)
                and not pa.types.is_int64(at)
                and not pa.types.is_float32(at)
                and not pa.types.is_int32(at)
            ):
                reasons.append(
                    f"Expected float64 or float32 or int64 or int32, got {at}"
                )
        elif isinstance(wt, types.Boolean):
            if not pa.types.is_boolean(at):
                reasons.append(f"Expected boolean, got {at}")
        elif isinstance(wt, types.NoneType):
            # We don't check for pa.types.is_null(at) because we can have for
            # example at == int32 with all nulls.
            if not arr.null_count == len(arr):
                reasons.append(f"Expected all nulls, got {arr.null_count}")
        elif isinstance(wt, types.Timestamp):
            if not pa.types.is_timestamp(at):
                reasons.append(f"Expected TimestampType, got {at}")
        elif types.is_custom_type(wt):
            if pa.types.is_dictionary(at):
                if not pa.types.is_string(at.value_type):
                    reasons.append(
                        f"Expected StringType as dictionary type, got {at.value_type}"
                    )
            elif not pa.types.is_string(at):
                reasons.append(f"Expected StringType, got {at}")
        else:
            reasons.append(f"Unhandled case: {wt}, {at}")

    if reasons:
        indented_reasons = textwrap.indent("\n".join(reasons), "  ")
        print("TYPE", type(wt))
        return f"{debug_types.short_type(wt)}, {str(at)[:50]}\n{indented_reasons}"

    return None


_awl_invalid_reason: contextvars.ContextVar[
    typing.Optional[str]
] = contextvars.ContextVar("_awl_invalid_reason", default=None)


@contextlib.contextmanager
def unsafe_awl_construction(reason: str):
    token = None
    if _awl_invalid_reason.get() is None:
        token = _awl_invalid_reason.set(reason)
    try:
        yield
    finally:
        if token is not None:
            _awl_invalid_reason.reset(token)


ArrowWeaveListObjectTypeVar = typing.TypeVar("ArrowWeaveListObjectTypeVar")


class ArrowWeaveList(typing.Generic[ArrowWeaveListObjectTypeVar]):
    _arrow_data: pa.Array
    object_type: types.Type

    # TODO: Refactor to disable None artifact? (Only used in tests)
    def __init__(
        self,
        arrow_data: typing.Union[pa.Table, pa.ChunkedArray, pa.Array],
        object_type=None,
        artifact: typing.Optional[artifact_base.Artifact] = None,
        invalid_reason=None,
    ) -> None:
        # Do not dictionary decode this array! That will break performance.
        # Note we combine chunks here, to make the internal interface easy
        # to use. In the future we could refactor to retain chunked form.
        if isinstance(arrow_data, pa.Array):
            self._arrow_data = arrow_data
        elif isinstance(arrow_data, pa.Table):
            self._arrow_data = pa.StructArray.from_arrays(
                # TODO: we shouldn't need to combine chunks, we can produce this in the
                # original chunked form for zero copy
                [c.combine_chunks() for c in arrow_data.columns],
                names=arrow_data.column_names,
            )
        elif isinstance(arrow_data, pa.ChunkedArray):
            self._arrow_data = arrow_data.combine_chunks()
        else:
            raise TypeError(
                "Expected pyarrow Array, ChunkdArray or Table, got %s"
                % type(arrow_data)
            )

        self.object_type = object_type
        if self.object_type is None:
            self.object_type = types.TypeRegistry.type_of(self._arrow_data).object_type
        self._artifact = artifact
        if invalid_reason is None:
            invalid_reason = _awl_invalid_reason.get()
        self._invalid_reason = invalid_reason

        self._validate()

    def _validate(self) -> None:
        if self._invalid_reason is None:
            self._validate_weave_type()
        self._validate_arrow_data()

    def _validate_weave_type(self) -> None:
        return

        # Validation disabled
        arr = self._arrow_data
        type_match_summary = weave_arrow_type_check(self.object_type, arr)
        if type_match_summary != None:
            print()
            print("ArrowWeaveList VALIDATION ERROR")
            print()
            print("ArrowWeaveList.object_type", self.object_type)
            print()
            print(
                "ArrowWeaveList._arrow_data.type\n",
                pretty_print_arrow_type(self._arrow_data.type),
            )
            print()
            print(type_match_summary)
            raise ValueError(f"ArrowWeaveList validation err: {type_match_summary}")

    def _validate_arrow_data(self) -> None:
        return

        # Validation disabled: this seems to usually be fast, but can take
        # many seconds for certain arrays.
        try:
            self._arrow_data.validate(full=True)
        except pa.ArrowInvalid:
            raise

    def validate(self) -> None:
        self._validate_weave_type()
        self._validate_arrow_data()
        self._invalid_reason = None

    def map_column(
        self,
        fn: typing.Callable[
            ["ArrowWeaveList", PathType], typing.Optional["ArrowWeaveList"]
        ],
        pre_fn: typing.Optional[
            typing.Callable[
                ["ArrowWeaveList", PathType], typing.Optional["ArrowWeaveList"]
            ]
        ] = None,
    ) -> "ArrowWeaveList":
        return self._map_column(fn, pre_fn, ())

    def _map_column(
        self,
        fn: typing.Callable[
            ["ArrowWeaveList", PathType], typing.Optional["ArrowWeaveList"]
        ],
        pre_fn: typing.Optional[
            typing.Callable[
                ["ArrowWeaveList", PathType], typing.Optional["ArrowWeaveList"]
            ]
        ],
        path: PathType,
    ) -> "ArrowWeaveList":
        if pre_fn is not None:
            pre_mapped = pre_fn(self, path)
            if pre_mapped is not None:
                self = pre_mapped
        with_mapped_children = self
        if isinstance(self.object_type, types.UnknownType):
            # Only occurs in empty list. Don't map into this.
            return self
        if isinstance(self.object_type, types.Const):
            with_mapped_children = ArrowWeaveList(
                self._arrow_data, self.object_type.val_type, self._artifact
            )._map_column(fn, pre_fn, path)
        elif isinstance(self.object_type, types.TypedDict):
            arr = self._arrow_data
            properties: dict[str, ArrowWeaveList] = {
                k: ArrowWeaveList(arr.field(k), v, self._artifact)._map_column(
                    fn, pre_fn, path + (PathItemStructField(k),)
                )
                for k, v in self.object_type.property_types.items()
            }
            if not properties:
                result_arr = arr
            else:
                result_arr = pa.StructArray.from_arrays(
                    [v._arrow_data for v in properties.values()],
                    list(properties.keys()),
                    mask=pa.compute.is_null(arr),
                )

            # set invalid_reason to the first non-None invalid reason found in
            # properties
            invalid_reason = first_non_none(
                v._invalid_reason for v in properties.values()
            )

            with_mapped_children = ArrowWeaveList(
                result_arr,
                types.TypedDict({k: v.object_type for k, v in properties.items()}),
                self._artifact,
                invalid_reason=invalid_reason,
            )
        elif isinstance(self.object_type, types.ObjectType):
            arr = self._arrow_data
            attrs: dict[str, ArrowWeaveList] = {
                k: ArrowWeaveList(arr.field(k), v, self._artifact)._map_column(
                    fn, pre_fn, path + (PathItemObjectField(k),)
                )
                for k, v in self.object_type.property_types().items()
            }

            # Types of some of the attrs may have changed. But some property types on object
            # types are no variable, and it is invalid to change them from our type system's
            # perspective. So the result here is invalid ArrowWeaveList. The results should
            # only be used in cirumstances where this is acceptable.
            with_mapped_children = ArrowWeaveList(
                pa.StructArray.from_arrays(
                    [v._arrow_data for v in attrs.values()],
                    list(attrs.keys()),
                    mask=pa.compute.is_null(arr),
                ),
                self.object_type,
                self._artifact,
                invalid_reason="mapped_column on ObjectType can produce invalid ArrowWeaveList",
            )
        elif isinstance(self.object_type, (types.List, ArrowWeaveListType)):
            arr = self._arrow_data
            items: ArrowWeaveList = ArrowWeaveList(
                arr.flatten(), self.object_type.object_type, self._artifact
            )._map_column(fn, pre_fn, path + (PathItemList(),))
            # print("SELF OBJECT TYPE", self.object_type)
            # print("SELF ARROW DATA TYPE", self._arrow_data.type)
            with_mapped_children = ArrowWeaveList(
                pa.ListArray.from_arrays(
                    offsets_starting_at_zero(self._arrow_data),
                    items._arrow_data,
                    mask=pa.compute.is_null(arr),
                ),
                self.object_type.__class__(items.object_type),
                self._artifact,
                invalid_reason=items._invalid_reason,
            )
        elif isinstance(self.object_type, tagged_value_type.TaggedValueType):
            arr = self._arrow_data
            tag: ArrowWeaveList = ArrowWeaveList(
                self._arrow_data.field("_tag"), self.object_type.tag, self._artifact
            )._map_column(fn, pre_fn, path + (PathItemTaggedValueTag(),))
            value: ArrowWeaveList = ArrowWeaveList(
                self._arrow_data.field("_value"), self.object_type.value, self._artifact
            )._map_column(fn, pre_fn, path + (PathItemTaggedValueValue(),))
            with_mapped_children = ArrowWeaveList(
                pa.StructArray.from_arrays(
                    [tag._arrow_data, value._arrow_data],
                    ["_tag", "_value"],
                    mask=pa.compute.is_null(arr),
                ),
                tagged_value_type.TaggedValueType(tag.object_type, value.object_type),  # type: ignore
                self._artifact,
                invalid_reason=tag._invalid_reason or value._invalid_reason,
            )
        elif isinstance(self.object_type, types.UnionType):
            non_none_members = [
                m for m in self.object_type.members if not isinstance(m, types.NoneType)
            ]
            nullable = len(non_none_members) < len(self.object_type.members)
            if len(non_none_members) == 1:
                non_none_member = ArrowWeaveList(
                    self._arrow_data,
                    non_none_members[0],
                    self._artifact,
                )._map_column(fn, pre_fn, path)
                with_mapped_children = ArrowWeaveList(
                    non_none_member._arrow_data,
                    types.optional(non_none_member.object_type),
                    self._artifact,
                    invalid_reason=non_none_member._invalid_reason,
                )
            elif len(non_none_members) > 1:
                arr = self._arrow_data
                members: list[ArrowWeaveList] = [
                    ArrowWeaveList(
                        arr.field(i),
                        types.optional(member_type) if nullable else member_type,
                        self._artifact,
                    )._map_column(fn, pre_fn, path + (PathItemUnionEntry(i),))
                    for i, member_type in enumerate(non_none_members)
                ]
                new_type_members = [m.object_type for m in members]
                if nullable:
                    new_type_members.append(types.NoneType())

                # I believe there is a bug here. We should merge the new members,
                # because they may have been re-written with new types, becoming
                # mergable. We don't seem to be hitting this problem in fuzz-testing
                # areas that make use of this function, I think because they either
                #   1. don't rewrite the types in a way that make them mergeable
                #      (e.g. to_pylists_tagged)
                #   2. do type rewriting but then don't rely on the weave types
                #      (e.g. to_compare_safe)
                #
                # However, I bet this will bite us!
                #
                # TODO: fix, by:
                # We'll need a "merge union" function that cleans up unions that
                # can be merged. The strategy is to try to merge pairs of types
                # using merge_types first. And in cases where that doesn't produce
                # a union, concat underlying member arrays together (using concat.py).

                # set invalid_reason to the first non-None invalid reason found in
                # properties
                invalid_reason = first_non_none(m._invalid_reason for m in members)
                with_mapped_children = ArrowWeaveList(
                    pa.UnionArray.from_dense(
                        self._arrow_data.type_codes,
                        self._arrow_data.offsets,
                        [m._arrow_data for m in members],
                    ),
                    # types.UnionType(*new_type_members),
                    types.union(*new_type_members),
                    self._artifact,
                    invalid_reason=invalid_reason,
                )
            else:
                raise errors.WeaveInternalError(
                    "Union must have at least one non-none member"
                )

        mapped = fn(with_mapped_children, path)
        if mapped is None:
            mapped = with_mapped_children
        return mapped

    def separate_tags(
        self,
    ) -> typing.Tuple["ArrowWeaveList", dict[PathType, "ArrowWeaveList"]]:
        tag_columns: dict[PathType, ArrowWeaveList] = {}

        def _remove_tags(list: ArrowWeaveList, path: PathType) -> ArrowWeaveList:
            if isinstance(list.object_type, tagged_value_type.TaggedValueType):
                # We need to remove _value components in the case that we are inside
                # tagged value here.
                tag_columns[path] = ArrowWeaveList(
                    list._arrow_data.field("_tag"), list.object_type.tag, list._artifact
                )
                return ArrowWeaveList(
                    list._arrow_data.field("_value"),
                    list.object_type.value,
                    list._artifact,
                )
            return list

        res = self.map_column(_remove_tags)

        reversed_tag_columns = {}
        for k, v in reversed(tag_columns.items()):
            reversed_tag_columns[k] = v

        return res, reversed_tag_columns

    def without_tags(self) -> "ArrowWeaveList":
        return self.separate_tags()[0]

    def separate_dictionaries(
        self,
    ) -> typing.Tuple["ArrowWeaveList", dict[PathType, "ArrowWeaveList"]]:
        dictionary_columns: dict[PathType, ArrowWeaveList] = {}

        def _remove_dictionaries(
            list: ArrowWeaveList, path: PathType
        ) -> typing.Optional[ArrowWeaveList]:
            if pa.types.is_dictionary(list._arrow_data.type):
                dictionary_columns[path] = ArrowWeaveList(
                    list._arrow_data.dictionary, list.object_type, list._artifact
                )
                return ArrowWeaveList(
                    list._arrow_data.indices,
                    # Don't change the object type to Int. Instead, we mark the ArrowWeaveList
                    # as invalid (the arrow type no longer matches object_type)
                    list.object_type,
                    list._artifact,
                    invalid_reason="Dictionary removed",
                )
            return None

        res = self.map_column(_remove_dictionaries), dictionary_columns
        return res

    def without_dictionaries(self) -> "ArrowWeaveList":
        def _remove_dictionaries(
            list: ArrowWeaveList, path: PathType
        ) -> typing.Optional[ArrowWeaveList]:
            if pa.types.is_dictionary(list._arrow_data.type):
                return ArrowWeaveList(
                    list._arrow_data.dictionary_decode(),
                    list.object_type,
                    list._artifact,
                    invalid_reason="Dictionary removed",
                )
            return None

        return self.map_column(_remove_dictionaries)

    def separate_awls(
        self,
    ) -> typing.Tuple[
        "ArrowWeaveList", dict[PathType, list[typing.Optional["ArrowWeaveList"]]]
    ]:
        awl_columns: dict[PathType, list[typing.Optional[ArrowWeaveList]]] = {}

        def _remove_awls(
            list: ArrowWeaveList, path: PathType
        ) -> typing.Optional[ArrowWeaveList]:
            if isinstance(list.object_type, ArrowWeaveListType):
                awl_columns[path] = [
                    ArrowWeaveList(
                        a.values, list.object_type.object_type, list._artifact
                    )
                    if a.values is not None
                    else None
                    for a in list._arrow_data
                ]
                return ArrowWeaveList(
                    pa.array([None] * len(list)), types.NoneType(), list._artifact
                )
            return None

        # Use pre-order traversal. We only want the top most awls.
        return self.map_column(lambda x, y: None, _remove_awls), awl_columns

    def function_type_paths(
        self,
    ) -> dict[PathType, types.Type]:
        function_type_paths: dict[PathType, types.Type] = {}

        def _save_obj_type_path(list: ArrowWeaveList, path: PathType) -> None:
            if isinstance(list.object_type, types.Function):
                function_type_paths[path] = list.object_type
            return None

        self.map_column(_save_obj_type_path)
        function_type_paths = reverse_dict(function_type_paths)
        return function_type_paths

    def obj_type_paths(
        self,
    ) -> dict[PathType, types.Type]:
        obj_type_paths: dict[PathType, types.Type] = {}

        def _save_obj_type_path(list: ArrowWeaveList, path: PathType) -> None:
            if isinstance(list.object_type, types.ObjectType):
                obj_type_paths[path] = list.object_type
            return None

        self.map_column(_save_obj_type_path)
        obj_type_paths = reverse_dict(obj_type_paths)
        return obj_type_paths

    def custom_type_paths(
        self,
    ) -> dict[PathType, types.Type]:
        custom_type_paths: dict[PathType, types.Type] = {}

        def _save_custom_type_path(list: ArrowWeaveList, path: PathType) -> None:
            if isinstance(
                list.object_type,
                # TODO: don't hardcode
                (
                    ArrowWeaveListType,
                    types.List,
                    types.Const,
                    types.Number,
                    types.Int,
                    types.Float,
                    types.Boolean,
                    types.String,
                    types.Timestamp,
                    types.NoneType,
                    tagged_value_type.TaggedValueType,
                    types.TypedDict,
                    types.ObjectType,
                    types.UnionType,
                    types.Function,
                ),
            ):
                return None
            custom_type_paths[path] = list.object_type

        self.map_column(_save_custom_type_path)
        return custom_type_paths

    def _arrow_data_asarray_no_tags(self) -> pa.Array:
        # Remove both tags and dictionaries. You should typically leave
        # dictionaries as late as possible! So only use this if you're sure
        # you don't want them.
        no_tags = self.without_tags()
        no_dicts = no_tags.without_dictionaries()
        return no_dicts._arrow_data

    def __array__(self, dtype=None):
        # TODO: replace with to_pylist_tagged once refs are supported
        pylist = self.to_pylist_raw()
        return np.asarray(pylist)

    def __iter__(self):
        pylist = self.to_pylist_tagged()
        for x in pylist:
            yield x

    def __repr__(self):
        return f"<ArrowWeaveList: {self.object_type}>"

    def to_pylist_raw(self):
        """Used for testing, preserves _tag and _value fields"""
        return self._to_pylist_dictsafe()

    def _to_pylist_dictsafe(self):
        value_awl, dict_columns = self.separate_dictionaries()
        value_py = value_awl._arrow_data.to_pylist()

        dict_columns = {p: c._arrow_data.to_pylist() for p, c in dict_columns.items()}

        # Dictionary decode the value
        if dict_columns:
            for path, dict_col in dict_columns.items():
                path = tuple(
                    p
                    if not isinstance(p, PathItemObjectField)
                    else PathItemStructField(p.attr)
                    for p in path
                )
                set_path(
                    value_py,
                    value_awl._arrow_data,
                    (PathItemOuterList(),) + path,
                    lambda v, j: None if v == None else dict_col[v],
                )

        return value_py

    def to_pylist_notags(self):
        """Convert the ArrowWeaveList to a python list, stripping tags"""
        value_awl, _ = self.separate_tags()
        return value_awl._to_pylist_dictsafe()

    def to_pylist_tagged(self):
        """Convert the ArrowWeaveList to a python list, tagging objects correctly"""
        value, awl_columns = self.separate_awls()

        custom_type_paths = value.custom_type_paths()
        function_type_paths = value.function_type_paths()
        obj_type_paths = value.obj_type_paths()

        value_awl, dict_columns = value.separate_dictionaries()
        value_py = value_awl._arrow_data.to_pylist()

        dict_columns = {p: c._arrow_data.to_pylist() for p, c in dict_columns.items()}

        for path, obj_type in obj_type_paths.items():
            # print("OBJ TYPE PATH", path)
            set_path(
                value_py,
                value_awl._arrow_data,
                (PathItemOuterList(),) + path,
                lambda v, j: None if v is None else obj_type.instance_class(**v),
            )
        # Dictionary decode the value, and add the tags to the tag store,
        # in a single pass.
        for path2, dict_col in dict_columns.items():
            set_path(
                value_py,
                value_awl._arrow_data,
                (PathItemOuterList(),) + path2,
                lambda v, j: None if v == None else dict_col[v],
            )

        for path, type_ in custom_type_paths.items():

            def custom_type_value(v, j):
                if v == None:
                    return None
                if isinstance(v, dict):
                    return type_.instance_from_dict(v)
                # else its a ref string
                # TODO: this does not use self.artifact, can we just drop it?
                # Do we need the type so we can load here? No...
                if ":" in v:
                    ref = ref_base.Ref.from_str(v)
                    # Note: we are forcing type here, because we already know it
                    # We don't save the types for every file in a remote artifact!
                    # But you can still reference them, because you have to get that
                    # file through an op, and therefore we know the type?
                    ref._type = type_
                    return ref.get()

                return self._artifact.get(v, type_)

            set_path(
                value_py,
                value_awl._arrow_data,
                (PathItemOuterList(),) + path,
                custom_type_value,
            )

        for path, function_type in function_type_paths.items():
            set_path(
                value_py,
                value_awl._arrow_data,
                (PathItemOuterList(),) + path,
                lambda v, j: node_ref.ref_to_node(ref_base.Ref.from_str(v)),
            )

        for path, awls in awl_columns.items():
            set_path(
                value_py,
                value_awl._arrow_data,
                (PathItemOuterList(),) + path,
                lambda v, j: awls[j],
            )
        return concrete_to_tagstore(value_py)

    def _count(self):
        return len(self._arrow_data)

    def __len__(self):
        return self._count()

    def replace_column(
        self,
        name: typing.Union[str, PathType],
        fn: typing.Callable[["ArrowWeaveList"], "ArrowWeaveList"],
    ):
        replaced = {"v": False}

        def _do_replace(
            list: ArrowWeaveList, path: PathType
        ) -> typing.Optional[ArrowWeaveList]:
            if (
                isinstance(name, tuple)
                and path == name
                or len(path) == 1
                and path[0] == name
            ):
                replaced["v"] = True
                return fn(list)
            return None

        res = self.map_column(_do_replace)
        if not replaced["v"]:
            raise ValueError("Column {} not found".format(name))
        return res

    def dictionary_encode(self) -> "ArrowWeaveList":
        return ArrowWeaveList(
            self._arrow_data.dictionary_encode(), self.object_type, self._artifact
        )

    def _index(
        self,
        index: typing.Optional[typing.Union[int, typing.List[typing.Optional[int]]]],
    ):
        if index == None:
            return None
        indexes: pa.Array
        if isinstance(index, int):
            indexes = [index]
        elif isinstance(index, ArrowWeaveList):
            indexes = index._arrow_data
        else:
            indexes = index

        arr = self._arrow_data
        length = len(arr)
        neg_cond = pc.less(indexes, 0)
        indexes_neg = pc.add(indexes, length)
        indexes = pc.choose(neg_cond, indexes, indexes_neg)
        out_of_bounds = pc.or_(pc.less(indexes, 0), pc.greater_equal(indexes, length))
        indexes_bounds_checked = pc.choose(
            out_of_bounds, indexes, pa.scalar(None, pa.int64())
        )
        result_rows = pc.take(arr, indexes_bounds_checked)
        awl: ArrowWeaveList = ArrowWeaveList(
            result_rows, self.object_type, self._artifact
        )
        if isinstance(index, int):
            return awl.to_pylist_tagged()[0]
        return awl

    def __getitem__(
        self,
        index: typing.Optional[typing.Union[int, typing.List[typing.Optional[int]]]],
    ):
        return self._index(index)

    def _slice(self, start: int, stop: int):
        # We don't slice, we take to get a copy. Arrow slicing is confusing.
        # I was seeing that type_codes in sliced unions were offset somehow...
        return ArrowWeaveList(
            self._arrow_data.take(np.arange(start, stop)),
            self.object_type,
            self._artifact,
        )

    def _append_column(self, name: str, data) -> "ArrowWeaveList":
        if not data:
            raise ValueError(f'Data for new column "{name}" must be nonnull.')

        if isinstance(self._arrow_data, pa.Table):
            new_data = self._arrow_data.append_column(name, [data])
        elif isinstance(self._arrow_data, pa.StructArray):
            chunked_arrays = {}
            for field in self._arrow_data.type:
                chunked_arrays[field.name] = pa.chunked_array(
                    self._arrow_data.field(field.name)
                )
            arrow_obj = pa.table(chunked_arrays)
            new_data = arrow_obj.append_column(name, [data])
        else:
            raise ValueError(
                f"Cannot append column to {type(self._arrow_data)} object."
            )

        return ArrowWeaveList(new_data, None, self._artifact)

    def keys(self) -> list[str]:
        if isinstance(self.object_type, tagged_value_type.TaggedValueType):
            value_type = self.object_type.value
            if isinstance(value_type, types.TypedDict):
                return list(value_type.property_types.keys())
        elif isinstance(self.object_type, types.TypedDict):
            return list(self.object_type.property_types.keys())
        raise ValueError("Cannot get keys from non-TypedDict.")

    def _split_none(self) -> typing.Tuple[bool, "ArrowWeaveList"]:
        if isinstance(self.object_type, types.UnionType):
            optional, non_none_type = types.split_none(self.object_type)
            return optional, ArrowWeaveList(
                self._arrow_data,
                non_none_type,
                self._artifact,
                invalid_reason="maybe optional",
            )
        return False, self

    def _as_optional(self) -> "ArrowWeaveList":
        invalid_reason = self._invalid_reason
        if self._invalid_reason == "maybe optional":
            invalid_reason = None
        return ArrowWeaveList(
            self._arrow_data,
            types.optional(self.object_type),
            self._artifact,
            invalid_reason=invalid_reason,
        )

    def _null_mask(self) -> pa.Array:
        return safe_is_null(self._arrow_data)

    def column(self, name: str) -> "ArrowWeaveList":
        if isinstance(self.object_type, tagged_value_type.TaggedValueType):
            return make_vec_taggedvalue(
                self.tagged_value_tag(),
                self.tagged_value_value().column(name),
                is_null_mask=self._null_mask(),
            )
        elif isinstance(self.object_type, types.UnionType):
            optional, non_none = self._split_none()
            if optional and not isinstance(non_none.object_type, types.UnionType):
                return non_none.column(name)._as_optional()
            raise ValueError(
                f"Cannot get column {name} from non-TypedDict or ObjectType."
            )
        elif isinstance(self.object_type, types.TypedDict):
            property_types = self.object_type.property_types
        elif isinstance(self.object_type, types.ObjectType):
            property_types = self.object_type.property_types()
        else:
            raise ValueError(
                f"Cannot get column {name} from non-TypedDict or ObjectType."
            )

        if name not in property_types:
            return make_vec_none(len(self))

        return ArrowWeaveList(
            self._arrow_data.field(name),
            property_types[name],
            self._artifact,
            invalid_reason=self._invalid_reason,
        )

    def tagged_value_tag(self) -> "ArrowWeaveListGeneric[types.TypedDict]":
        if not isinstance(self.object_type, tagged_value_type.TaggedValueType):
            raise ValueError(f"Cannot get tagged_value_tag from non-TaggedValueType")

        return ArrowWeaveListGeneric[types.TypedDict](
            self._arrow_data.field("_tag"), self.object_type.tag, self._artifact
        )

    def tagged_value_value(self) -> "ArrowWeaveList":
        if not isinstance(self.object_type, tagged_value_type.TaggedValueType):
            raise ValueError(f"Cannot get tagged_value_tag from non-TaggedValueType")

        return ArrowWeaveList(
            self._arrow_data.field("_value"), self.object_type.value, self._artifact
        )

    def untagged(self) -> "ArrowWeaveList":
        if isinstance(self.object_type, tagged_value_type.TaggedValueType):
            return self.tagged_value_value()
        return self

    def _make_lambda_node(
        self, fn: typing.Union[typing.Callable[[typing.Any], typing.Any], graph.Node]
    ):
        if isinstance(fn, graph.Node):
            return fn

        sig = inspect.signature(fn)
        if len(sig.parameters) == 1:
            vars = {"row": self.object_type}
        elif len(sig.parameters) == 2:
            vars = {"row": self.object_type, "index": types.Int()}
        else:
            raise ValueError(
                "Functions passed to ArrowWeaveList.apply must have 1 or 2 parameters (row, [index])"
            )
        return weave_internal.define_fn(vars, fn).val

    def apply(
        self, fn: typing.Union[typing.Callable[[typing.Any], typing.Any], graph.Node]
    ):
        fn = self._make_lambda_node(fn)
        from .vectorize import _apply_fn_node_with_tag_pushdown

        return _apply_fn_node_with_tag_pushdown(self, fn)  # type: ignore

    def concat(self, other: "ArrowWeaveList") -> "ArrowWeaveList":
        from . import concat

        return concat.concatenate(self, other)

    def join2(
        self: "ArrowWeaveList",
        other: "ArrowWeaveList",
        join1Fn: graph.OutputNode,
        join2Fn: graph.OutputNode,
        alias1: typing.Optional[str] = None,
        alias2: typing.Optional[str] = None,
        leftOuter: bool = False,
        rightOuter: bool = False,
    ):
        join1Fn = self._make_lambda_node(join1Fn)
        join2Fn = other._make_lambda_node(join2Fn)
        from . import list_join

        return list_join.join2_impl(
            self, other, join1Fn, join2Fn, alias1, alias2, leftOuter, rightOuter
        )

    def _limit(self, limit: int):
        return ArrowWeaveList(
            self._arrow_data.slice(0, limit), self.object_type, self._artifact
        )

    def _with_object_type(self, object_type: types.Type, invalid_reason=None):
        return ArrowWeaveList(
            self._arrow_data, object_type, self._artifact, invalid_reason=invalid_reason
        )

    def _clear_invalid_reason(self):
        return ArrowWeaveList(
            self._arrow_data, self.object_type, self._artifact, invalid_reason=None
        )


ArrowWeaveListGenericType = typing.TypeVar(
    "ArrowWeaveListGenericType", bound=types.Type
)


class ArrowWeaveListGeneric(
    ArrowWeaveList[typing.Any], typing.Generic[ArrowWeaveListGenericType]
):
    object_type: ArrowWeaveListGenericType


ArrowWeaveListType.instance_classes = ArrowWeaveList


ArrowWeaveListType.instance_classes = ArrowWeaveList
ArrowWeaveListType.instance_class = ArrowWeaveList


def is_typedict_arrowweavelist(
    val: ArrowWeaveList,
) -> typing_extensions.TypeGuard[ArrowWeaveListGeneric[types.TypedDict]]:
    return isinstance(val.object_type, types.TypedDict)


def is_object_arrowweavelist(
    val: ArrowWeaveList,
) -> typing_extensions.TypeGuard[ArrowWeaveListGeneric[types.ObjectType]]:
    return isinstance(val.object_type, types.ObjectType)


def is_taggedvalue_arrowweavelist(
    val: ArrowWeaveList,
) -> typing_extensions.TypeGuard[
    ArrowWeaveListGeneric[tagged_value_type.TaggedValueType]
]:
    return isinstance(val.object_type, tagged_value_type.TaggedValueType)


def is_list_arrowweavelist(
    val: ArrowWeaveList,
) -> typing_extensions.TypeGuard[ArrowWeaveListGeneric[types.List]]:
    return isinstance(val.object_type, types.List)


def dataframe_to_arrow(df):
    return ArrowWeaveList(pa.Table.from_pandas(df))


pandas_to_awl = dataframe_to_arrow


def make_vec_none(length: int) -> ArrowWeaveList:
    return ArrowWeaveList(pa.nulls(length), types.NoneType(), None)


def make_vec_dict(**kwargs: ArrowWeaveList):
    arr = pa.StructArray.from_arrays(
        [v._arrow_data for v in kwargs.values()], [k for k in kwargs.keys()]
    )
    property_types = {k: v.object_type for k, v in kwargs.items()}
    return ArrowWeaveList(arr, types.TypedDict(property_types), None)


def make_vec_taggedvalue(
    tag: ArrowWeaveListGeneric[types.TypedDict],
    value: ArrowWeaveList,
    is_null_mask: typing.Optional[pa.Array] = None,
):
    tag_types = tag.object_type.property_types
    value_type = value.object_type
    tag_data = tag._arrow_data
    value_data = value._arrow_data
    if isinstance(value_type, tagged_value_type.TaggedValueType):
        # If value is a tagged value, we flatten by merging the tags and
        # pulling the value up.
        value_type_types = value_type.tag.property_types
        new_tag_types = {**tag_types, **value_type_types}
        if len(new_tag_types) != len(tag_types) + len(value_type_types):
            raise ValueError("Tagged value types must be disjoint")
        tag_types = new_tag_types
        value_type = value_type.value

        value_tag_data = value_data.field("_tag")
        tag_data = pa.StructArray.from_arrays(
            [tag_data.field(f.name) for f in tag_data.type]
            + [value_tag_data.field(f.name) for f in value_tag_data.type],
            names=[f.name for f in tag_data.type]
            + [f.name for f in value_tag_data.type],
        )
        value_data = value_data.field("_value")
    arr = pa.StructArray.from_arrays(
        [tag_data, value_data],
        ["_tag", "_value"],
        mask=is_null_mask,
    )
    return ArrowWeaveList(
        arr, tagged_value_type.TaggedValueType(types.TypedDict(tag_types), value_type), None  # type: ignore
    )


def awl_zip(*arrs: ArrowWeaveList) -> ArrowWeaveList:
    if not arrs:
        raise ValueError("Cannot zip empty list")
    from . import convert

    arrs = convert.unify_types(*arrs)
    zipped = arrow_zip(*[a._arrow_data for a in arrs])
    return ArrowWeaveList(zipped, types.List(arrs[0].object_type), None)



import dataclasses
from tarfile import DEFAULT_FORMAT
import typing
import inspect
from . import api as weave
from . import graph
from . import storage
from . import weave_types as types
from . import weave_internal
from . import panel_util
from . import errors
from .ops import get as op_get


# class PanelType(types.Type):
#     name = "panel"

#     def instance_to_dict(self, obj):
#         res = {
#             "id": obj.id,
#             "input": weavejs_fixes.fixup_node(obj.input_node).to_json(),
#             "config": weavejs_fixes.fixup_data(obj.config),
#         }
#         return res

#     def instance_from_dict(self, d):
#         obj = Panel(graph.Node.node_from_json(d["input"]))
#         obj.id = d["id"]
#         obj.config = d["config"]
#         return obj


def run_variable_lambdas(obj, vars):
    if not isinstance(obj, graph.Node) and callable(obj):
        sig = inspect.signature(obj)
        param_nodes = {}
        for param in sig.parameters:
            try:
                param_type = vars[param].type
            except KeyError:
                raise errors.WeaveDefinitionError(
                    "Variable '%s' not available in frame: %s" % (param, vars)
                )
            if isinstance(param_type, types.Function):
                param_type = param_type.output_type
            param_nodes[param] = weave_internal.make_var_node(param_type, param)
        injected = obj(**param_nodes)
        return run_variable_lambdas(injected, vars)
    if isinstance(obj, list):
        return [run_variable_lambdas(v, vars) for v in obj]
    elif isinstance(obj, dict):
        return {k: run_variable_lambdas(v, vars) for k, v in obj.items()}
    else:
        return obj


# So this needs to take:
# user can pass
#   - input
#   - user assignments
#   - settings (initial state, not 1to1 with config)
# internal
#   - config
#   - scope assignments


# @weave.weave_class(weave_type=PanelType)
# @dataclasses.dataclass


class ConfigDescriptor:
    def __set_name__(self, owner, name):
        self._name = "_" + name

    def __get__(self, obj, type):
        try:
            fields = dataclasses.fields(obj)
        except TypeError:
            return {}
        conf = {}
        for field in fields:
            if (
                field.name == "input_node"
                or field.name == "vars"
                or field.name == "config"
            ):
                continue
            conf[field.name] = getattr(obj, field.name)
        return conf

    def __set__(self, obj, value):
        pass
        # setattr(obj, self._name, int(value))


InputNodeType = typing.TypeVar("InputNodeType")
VarsType = typing.TypeVar("VarsType")


# Currently, storage.save() works on Panels, but storage.get()
# does not.
@weave.type(__init=False)  # type: ignore
class Panel(typing.Generic[InputNodeType, VarsType]):
    id: str = dataclasses.field(init=False)
    input_node: graph.Node = dataclasses.field(default=graph.VoidNode())
    vars: dict[str, graph.Node] = dataclasses.field(default_factory=dict)

    def __init__(
        self, input_node=None, vars=None, config=None, _renderAsPanel=None, **options
    ):
        self.vars = {}
        if vars:
            for name, val in vars.items():
                self.vars[name] = panel_util.make_node(val)

        self.input_node = input_node
        self.input_node = run_variable_lambdas(self.input_node, self.vars)

        if not isinstance(self.input_node, graph.Node):
            self.input_node = storage.to_safe_const(self.input_node)

        self.config = config
        if config is None and self.__init__.__func__ == Panel.__init__:
            # Auto initialize if the child doesn't override __init__
            config_anno = self.__annotations__.get("config")
            if config_anno is not None:
                if hasattr(self, "initialize"):
                    self.config = weave.use(self.initialize())
                else:
                    if typing.get_origin(config_anno) is typing.Union:
                        config_class = typing.get_args(config_anno)[0]
                    else:
                        config_class = config_anno
                    self.config = config_class()

        for k, v in options.items():
            config_val = getattr(self.config, k)
            if config_val is None:
                raise TypeError(f'Panel {self.id} has not option "{k}"')
            if isinstance(config_val, graph.Node) and isinstance(
                config_val.type, types.Function
            ):
                new_config_val = weave.define_fn(config_val.type.input_types, v)
            elif isinstance(config_val, graph.Node):
                new_config_val = panel_util.make_node(v)
            else:
                new_config_val = v
            setattr(self.config, k, new_config_val)

        rendered_panel_anno = self.__annotations__.get("_renderAsPanel")
        if rendered_panel_anno is not None:
            if _renderAsPanel:
                self._renderAsPanel = _renderAsPanel
            else:
                if typing.get_origin(rendered_panel_anno) is typing.Union:
                    render_as_panel_class = typing.get_args(rendered_panel_anno)[0]
                else:
                    render_as_panel_class = rendered_panel_anno
                self._renderAsPanel = render_as_panel_class()

    def _normalize(self, frame=None):
        pass

    def to_json(self):
        return {
            "vars": {k: v.to_json() for k, v in self.vars.items()},
            "input_node": self.input_node.to_json(),
            "id": self.id,
            "config": self.config,
        }



from weave import context_state

loading_builtins_token = context_state.set_loading_built_ins()

try:
    from . import example

    from . import bertviz

    from . import xgboost
    from . import shap

    from . import sklearn
    from . import torchvision
    from . import torch_mnist_model_example
    from . import huggingface

    from . import craiyon

    from . import spacy
    from . import lens
    from . import wandb
    from . import scenario
    from . import shawn
    from . import wandb
    from . import replicate
    from . import openai
    from . import py

    from . import langchain

    from . import umap
    from . import hdbscan

finally:
    context_state.clear_loading_built_ins(loading_builtins_token)



"""Fixes for compatibility with WeaveJS

TODO: this file is not complete. We should try to put all compability fixes here.
    Grep for "js_compat" to find other instances.
"""
import typing
import copy
import math

from . import weave_types
from . import graph
from . import weave_types
from . import errors


def _convert_specific_opname_to_generic_opname(
    name: str, inputs: dict[str, typing.Any]
) -> tuple[str, dict[str, typing.Any]]:
    if (
        name == "typedDict-pick"
        or name == "dict-pick"
        or name == "list-pick"
        or name == "ArrowWeaveListTypedDict-pick"
    ):
        return "pick", {"obj": inputs["self"], "key": inputs["key"]}
    elif name == "ArrowWeaveList-concat":
        return "concat", {"arr": inputs["arr"]}
    elif name == "groupresult-map":
        return "map", {"arr": inputs["self"], "mapFn": inputs["map_fn"]}
    elif name == "groupresult-groupby":
        return "groupby", {"arr": inputs["self"], "groupByFn": inputs["group_fn"]}
    elif name == "groupresult-count" or name == "projectArtifactVersions-count":
        return "count", {"arr": inputs["self"]}
    elif name == "groupresult-key":
        return "group-groupkey", {"obj": inputs["self"]}
    elif name == "list-map":
        return "map", {"arr": inputs["self"], "mapFn": inputs["map_fn"]}
    elif name == "list-groupby":
        return "groupby", {"arr": inputs["self"], "groupByFn": inputs["group_by_fn"]}
    elif name == "list-filter" or name == "ArrowWeaveList-filter":
        return "filter", {"arr": inputs["self"], "filterFn": inputs["filter_fn"]}
    elif name == "list-__getitem__":
        return "index", {"arr": inputs["arr"], "index": inputs["index"]}
    elif name == "ArrowWeaveList-limit":
        return "limit", {"arr": inputs["self"], "limit": inputs["limit"]}
    elif name == "ArrowWeaveList-map":
        return "map", {"arr": inputs["self"], "mapFn": inputs["map_fn"]}
    elif name == "ArrowWeaveListNumber-min":
        return "numbers-min", {"numbers": inputs["self"]}
    elif name == "ArrowWeaveListNumber-max":
        return "numbers-max", {"numbers": inputs["self"]}
    elif name == "ArrowWeaveListDate-min":
        return "dates-min", {"dates": inputs["self"]}
    elif name == "ArrowWeaveListDate-max":
        return "dates-max", {"dates": inputs["self"]}
    elif (
        name == "groupresult-__getitem__"
        or name == "artifacts-__getitem__"
        or name == "projectArtifactVersions-__getitem__"
        or name == "ArrowWeaveList-__getitem__"
    ):
        return "index", {"arr": inputs["self"], "index": inputs["index"]}
    return name, inputs


def convert_specific_opname_to_generic_opname(
    name: str, inputs: dict[str, typing.Any]
) -> tuple[str, dict[str, typing.Any]]:
    if name.startswith("mapped_"):
        unmapped_name = name[7:]
        return _convert_specific_opname_to_generic_opname(unmapped_name, inputs)
    return _convert_specific_opname_to_generic_opname(name, inputs)


def convert_specific_ops_to_generic_ops_node(node: graph.Node) -> graph.Node:
    """Converts specific ops like typedDict-pick to generic ops like pick"""

    def convert_specific_op_to_generic_op(node: graph.Node):
        if isinstance(node, graph.ConstNode) and isinstance(
            node.type, weave_types.Function
        ):
            return graph.ConstNode(
                node.type, convert_specific_ops_to_generic_ops_node(node.val)
            )
        if not isinstance(node, graph.OutputNode):
            return node
        name, inputs = convert_specific_opname_to_generic_opname(
            node.from_op.name, node.from_op.inputs
        )
        return graph.OutputNode(node.type, name, inputs)

    return graph.map_nodes_full([node], convert_specific_op_to_generic_op)[0]


def _obj_is_node_like(data: typing.Any):
    if not isinstance(data, dict):
        return False
    return data.get("nodeType") in ["const", "output", "var", "void"]


# Non-perfect heuristic to determine if a serialized dict is likely an op
def _dict_is_op_like(data: dict):
    # Firstly, ops will only have "name" and "input" keys
    if set(data.keys()) == set(["name", "inputs"]):
        # Those keys will be str and list respectively.
        if isinstance(data["name"], str) and isinstance(data["inputs"], dict):
            # And the inputs will be dicts that are node-like
            return all(
                _obj_is_node_like(in_node) for in_node in data["inputs"].values()
            )
    return False


def convert_specific_ops_to_generic_ops_data(data):
    """Fix op call names for serialized objects containing graphs"""
    if isinstance(data, list):
        return [convert_specific_ops_to_generic_ops_data(d) for d in data]
    elif isinstance(data, dict):
        d = data
        if _dict_is_op_like(data):
            d["name"], d["inputs"] = convert_specific_opname_to_generic_opname(
                d["name"], d["inputs"]
            )
        return {k: convert_specific_ops_to_generic_ops_data(v) for k, v in d.items()}
    return data


def remove_opcall_versions_node(node: graph.Node) -> graph.Node:
    """Fix op call names"""

    def remove_op_version(node: graph.Node):
        if not isinstance(node, graph.OutputNode):
            return node
        return graph.OutputNode(
            node.type,
            graph.op_full_name(node.from_op),
            node.from_op.inputs,
        )

    return graph.map_nodes_full([node], remove_op_version)[0]


def remove_opcall_versions_data(data):
    """Fix op call names for serialized objects containing graphs"""
    if isinstance(data, list):
        return [remove_opcall_versions_data(d) for d in data]
    elif isinstance(data, dict):
        d = data
        if _dict_is_op_like(data):
            d = copy.copy(data)
            d["name"] = graph.opuri_full_name(data["name"])
        return {k: remove_opcall_versions_data(v) for k, v in d.items()}
    return data


def fixup_node(node: graph.Node) -> graph.Node:
    node = remove_opcall_versions_node(node)
    return convert_specific_ops_to_generic_ops_node(node)


def recursively_unwrap_unions(obj):
    if isinstance(obj, list):
        return [recursively_unwrap_unions(o) for o in obj]
    if isinstance(obj, dict):
        if "_union_id" in obj and "_val" in obj:
            return recursively_unwrap_unions(obj["_val"])
        else:
            return {
                k: recursively_unwrap_unions(v)
                for k, v in obj.items()
                if k != "_union_id"
            }
    return obj


def remove_nan_and_inf(obj):
    if isinstance(obj, list):
        return [remove_nan_and_inf(o) for o in obj]
    if isinstance(obj, dict):
        return {k: remove_nan_and_inf(v) for k, v in obj.items()}
    if isinstance(obj, float):
        if math.isnan(obj) or math.isinf(obj):
            return 0
    return obj


def fixup_data(data):
    data = recursively_unwrap_unions(data)
    data = remove_opcall_versions_data(data)
    # No good! We have to do this because remoteHttp doesn't handle NaN/inf in
    # response right now.
    # TODO: fix. Encode as string and then interpret in js side.
    data = remove_nan_and_inf(data)
    return convert_specific_ops_to_generic_ops_data(data)



import functools
import json
import typing

from . import errors
from . import weave_types
from . import uris
from . import storage

T = typing.TypeVar("T")


# The Generic here is currently only used in op definitions, its not actually
# applied within the class. See infer_types.py for usage.
class Node(typing.Generic[T]):
    type: weave_types.Type

    def __deepcopy__(self, memo: dict) -> "Node":
        return self.node_from_json(self.to_json())

    @classmethod
    def node_from_json(cls, obj: dict) -> "Node":
        if obj["nodeType"] == "const":
            return ConstNode.from_json(obj)
        elif obj["nodeType"] == "output":
            return OutputNode.from_json(obj)
        elif obj["nodeType"] == "var":
            return VarNode.from_json(obj)
        elif obj["nodeType"] == "void":
            return VoidNode()
        else:
            raise errors.WeaveInternalError("invalid node type: %s" % obj)

    def to_json(self) -> dict:
        raise NotImplementedError

    def __hash__(self) -> int:
        # We store nodes in a memoize cache in execute.py. They need to be
        # hashable. But the number.py ops override __eq__ which makes the default
        # Python hash not work, so we fix it up here.
        return id(self)

    def __str__(self) -> str:
        return node_expr_str(self)

    def __repr__(self) -> str:
        return "<%s(%s): %s %s>" % (
            self.__class__.__name__,
            id(self),
            self.type,
            str(self),
        )

    def __bool__(self) -> bool:
        raise errors.WeaveTypeError("Cannot use a node as a boolean predicate.")


# weave_types.Function.instance_classes = Node
weave_types.Function.instance_classes.append(Node)

OpInputNodeT = typing.TypeVar("OpInputNodeT", bound=Node)

Frame = typing.Mapping[str, Node]


class Op(typing.Generic[OpInputNodeT]):
    name: str
    inputs: dict[str, OpInputNodeT]
    _full_name: typing.Optional[str]

    def __init__(self, name: str, inputs: dict[str, OpInputNodeT]) -> None:
        # TODO: refactor this variable to be "uri"
        self.name = name
        self.inputs = inputs
        self._full_name = None

    @property
    def friendly_name(self) -> str:
        full_name = self.full_name
        return full_name.split("-", 1)[-1]

    # Called frequently and expensive, so cache it.
    @functools.cached_property
    def full_name(self) -> str:
        return opuri_full_name(self.name)

    # Called frequently and expensive, so cache it.
    @functools.cached_property
    def input_types(self) -> dict[str, weave_types.Type]:
        its: dict[str, weave_types.Type] = {}
        for k, v in self.inputs.items():
            if isinstance(v, ConstNode):
                its[k] = weave_types.Const(v.type, v.val)
            else:
                its[k] = v.type
        return its

    def to_json(self) -> dict:
        json_inputs = {}
        for k, v in self.inputs.items():
            json_inputs[k] = v.to_json()
        return {"name": self.name, "inputs": json_inputs}

    def __repr__(self) -> str:
        return f"<Op({id(self)} name={self.name} inputs={self.inputs}>"


class OutputNode(Node, typing.Generic[OpInputNodeT]):
    from_op: Op[OpInputNodeT]
    val: typing.Any

    def __init__(
        self, type: weave_types.Type, op_name: str, op_inputs: dict[str, OpInputNodeT]
    ) -> None:
        self.type = type
        self.from_op = Op(op_name, op_inputs)

    @classmethod
    def from_json(cls, val: dict) -> "OutputNode":
        op_inputs = val["fromOp"]["inputs"]
        inputs: dict[str, OpInputNodeT] = {}
        for param_name, param_node_json in op_inputs.items():
            # I am not sure why we need a cast here, `OpInputNodeT` is bound to
            # `Node` so it should be fine.
            inputs[param_name] = typing.cast(
                OpInputNodeT, Node.node_from_json(param_node_json)
            )
        return cls(
            weave_types.TypeRegistry.type_from_dict(val["type"]),
            val["fromOp"]["name"],
            inputs,
        )

    def iteritems_op_inputs(self) -> typing.Iterator[typing.Tuple[str, Node]]:
        return iter(self.from_op.inputs.items())

    def to_json(self) -> dict:
        return {
            "nodeType": "output",
            "type": self.type.to_dict(),
            "fromOp": self.from_op.to_json(),
        }

    def __repr__(self) -> str:
        return "<OutputNode(%s) type: %s op_name: %s>" % (
            id(self),
            self.type,
            self.from_op.name,
        )


class VarNode(Node):
    name: str

    def __init__(self, type: weave_types.Type, name: str) -> None:
        self.type = type
        self.name = name

    @classmethod
    def from_json(cls, val: dict) -> "VarNode":
        return cls(weave_types.TypeRegistry.type_from_dict(val["type"]), val["varName"])

    def to_json(self) -> dict:
        return {"nodeType": "var", "type": self.type.to_dict(), "varName": self.name}


class ConstNode(Node):
    val: typing.Any

    def __init__(self, type: weave_types.Type, val: typing.Any) -> None:
        self.type = type
        self.val = val

    @classmethod
    def from_json(cls, obj: dict) -> "ConstNode":
        from . import dispatch

        val = obj["val"]
        if isinstance(val, dict) and "nodeType" in val:
            val = Node.node_from_json(val)
        else:
            val = storage.from_python({"_type": obj["type"], "_val": obj["val"]})  # type: ignore
        t = weave_types.TypeRegistry.type_from_dict(obj["type"])
        if isinstance(t, weave_types.Function):
            cls = dispatch.RuntimeConstNode
        return cls(t, val)

    def to_json(self) -> dict:
        val = storage.to_python(self.val)["_val"]  # type: ignore
        # mapper = mappers_python.map_to_python(self.type, None)
        # val = mapper.apply(self.val)

        # val = self.val
        # if isinstance(self.type, weave_types.Function):
        #     val = val.to_json()
        return {"nodeType": "const", "type": self.type.to_dict(), "val": val}


class VoidNode(Node):
    type = weave_types.Invalid()

    def to_json(self) -> dict:
        return {"nodeType": "void", "type": "invalid"}


def nodes_equal(n1: Node, n2: Node) -> bool:
    return n1.to_json() == n2.to_json()


def for_each(graph: Node, visitor: typing.Callable[[Node], None]) -> None:
    if isinstance(graph, OutputNode):
        for param_name, param_node in graph.from_op.inputs.items():
            for_each(param_node, visitor)
    visitor(graph)


def opuri_full_name(op_uri: str) -> str:
    uri = uris.WeaveURI.parse(op_uri)
    return uri.name


def op_full_name(op: Op) -> str:
    return opuri_full_name(op.name)


def node_expr_str(node: Node) -> str:
    if isinstance(node, OutputNode):
        param_names = list(node.from_op.inputs.keys())
        if node.from_op.name == "dict":
            return "{%s}" % ", ".join(
                (
                    '"%s": %s' % (k, node_expr_str(n))
                    for k, n in node.from_op.inputs.items()
                )
            )
        elif node.from_op.name.endswith("__getattr__"):
            inputs = list(node.from_op.inputs.values())
            return "%s.%s" % (
                node_expr_str(inputs[0]),
                inputs[1].val,
            )
        elif node.from_op.name.endswith("pick") or node.from_op.name.endswith(
            "__getitem__"
        ):
            inputs = list(node.from_op.inputs.values())
            return "%s[%s]" % (
                node_expr_str(inputs[0]),
                node_expr_str(inputs[1]),
            )
        elif node.from_op.name == "gqlroot-wbgqlquery":
            query_hash = "_query_"  # TODO: make a hash from the query for idenity
            return f"{node.from_op.friendly_name}({query_hash})"
        elif all([not isinstance(n, OutputNode) for n in node.from_op.inputs.values()]):
            return "%s(%s)" % (
                node.from_op.friendly_name,
                ", ".join(node_expr_str(node.from_op.inputs[n]) for n in param_names),
            )
        if not param_names:
            return "%s()" % node.from_op.friendly_name
        return "%s.%s(%s)" % (
            node_expr_str(node.from_op.inputs[param_names[0]]),
            node.from_op.friendly_name,
            ", ".join(node_expr_str(node.from_op.inputs[n]) for n in param_names[1:]),
        )
    elif isinstance(node, ConstNode):
        if isinstance(node.type, weave_types.Function):
            res = node_expr_str(node.val)
            return res
        try:
            return json.dumps(node.val)
        except TypeError:
            # WARNING: This behavior means that sometimes this function
            # produces expressionions that JS can't parse (it happens when
            # we have Python Objects as values that have not yet been serialized)
            # TODO: fix
            return str(node.val)
    elif isinstance(node, VarNode):
        return node.name
    elif isinstance(node, VoidNode):
        return "<void>"
    else:
        return "**PARSE_ERROR**"


def _map_nodes(
    node: Node,
    map_fn: typing.Callable[[Node], typing.Optional[Node]],
    already_mapped: dict[Node, Node],
    walk_lambdas: bool,
) -> Node:
    # This is an iterative implemenation, to avoid blowing the stack and
    # to provide friendlier stack traces for exception merging tools.
    to_consider = [node]
    while to_consider:
        curr_node = to_consider[-1]
        if curr_node in already_mapped:
            to_consider.pop()
            continue
        result_node = curr_node
        if isinstance(curr_node, OutputNode):
            inputs = {}
            need_inputs = False
            for param_name, param_node in curr_node.from_op.inputs.items():
                if param_node not in already_mapped:
                    to_consider.append(param_node)
                    need_inputs = True
                else:
                    inputs[param_name] = already_mapped[param_node]
            if need_inputs:
                continue
            if any(n is not inputs[k] for k, n in curr_node.from_op.inputs.items()):
                result_node = OutputNode(curr_node.type, curr_node.from_op.name, inputs)
        elif (
            walk_lambdas
            and isinstance(curr_node, ConstNode)
            and isinstance(curr_node.type, weave_types.Function)
        ):
            if curr_node.val not in already_mapped:
                to_consider.append(curr_node.val)
                continue
            if curr_node.val is not already_mapped[curr_node.val]:
                result_node = ConstNode(curr_node.type, already_mapped[curr_node.val])

        to_consider.pop()
        mapped_node = map_fn(result_node)
        if mapped_node is None:
            mapped_node = result_node
        already_mapped[curr_node] = mapped_node
    return already_mapped[node]


OnErrorFnType = typing.Optional[typing.Callable[[int, Exception], Node]]


def map_nodes_top_level(
    leaf_nodes: list[Node],
    map_fn: typing.Callable[[Node], typing.Optional[Node]],
    on_error: OnErrorFnType = None,
) -> list[Node]:
    """Map nodes in dag represented by leaf nodes, but not sub-lambdas"""
    already_mapped: dict[Node, Node] = {}
    results: list[Node] = []
    for node_ndx, node in enumerate(leaf_nodes):
        try:
            results.append(_map_nodes(node, map_fn, already_mapped, False))
        except Exception as e:
            if on_error:
                results.append(on_error(node_ndx, e))
            else:
                raise e

    return results


def map_nodes_full(
    leaf_nodes: list[Node],
    map_fn: typing.Callable[[Node], typing.Optional[Node]],
    on_error: OnErrorFnType = None,
) -> list[Node]:
    """Map nodes in dag represented by leaf nodes, including sub-lambdas"""
    already_mapped: dict[Node, Node] = {}
    results: list[Node] = []
    for node_ndx, node in enumerate(leaf_nodes):
        try:
            results.append(_map_nodes(node, map_fn, already_mapped, True))
        except Exception as e:
            if on_error:
                results.append(on_error(node_ndx, e))
            else:
                raise e

    return results


def all_nodes_full(leaf_nodes: list[Node]) -> list[Node]:
    result: list[Node] = []
    map_nodes_full(leaf_nodes, lambda n: result.append(n))
    return result


def filter_nodes_top_level(
    nodes: list[Node], filter_fn: typing.Callable[[Node], bool]
) -> list[Node]:
    """Filter nodes in dag represented by leaf nodes, but not sub-lambdas"""
    result = []

    def mapped_fn(node: Node) -> Node:
        if filter_fn(node):
            result.append(node)
        return node

    map_nodes_top_level(nodes, mapped_fn)
    return result


def filter_nodes_full(
    nodes: list[Node], filter_fn: typing.Callable[[Node], bool]
) -> list[Node]:
    """Filter nodes in dag represented by leaf nodes, including sub-lambdas"""
    result = []

    def mapped_fn(node: Node) -> Node:
        if filter_fn(node):
            result.append(node)
        return node

    map_nodes_full(nodes, mapped_fn)
    return result


def expr_vars(node: Node) -> list[VarNode]:
    return typing.cast(
        list[VarNode], filter_nodes_top_level([node], lambda n: isinstance(n, VarNode))
    )


def is_open(node: Node) -> bool:
    """A Node is 'open' (as in open function) if there are one or more VarNodes"""
    return len(filter_nodes_top_level([node], lambda n: isinstance(n, VarNode))) > 0


def _all_nodes(node: Node) -> set[Node]:
    if not isinstance(node, OutputNode):
        return set((node,))
    res: set[Node] = set((node,))
    for input in node.from_op.inputs.values():
        res.update(_all_nodes(input))
    return res


def count(node: Node) -> int:
    return len(_all_nodes(node))


def _linearize(node: OutputNode) -> list[OutputNode]:
    if not node.from_op.inputs:
        return [node]
    arg0 = next(iter(node.from_op.inputs.values()))
    if not isinstance(arg0, OutputNode):
        return [node]
    return _linearize(arg0) + [node]


def linearize(node: Node) -> typing.Optional[list[OutputNode]]:
    """Return a list of the nodes by walking 0th argument."""
    if not isinstance(node, OutputNode):
        return None
    return _linearize(node)


def map_const_nodes_to_x(node: Node) -> Node:
    return map_nodes_full(
        [node],
        lambda n: n if not isinstance(n, ConstNode) else VarNode(n.type, "x"),
    )[0]



import json
import pathlib

import weave


def dirsize(path) -> int:
    return sum(f.stat().st_size for f in path.glob("**/*") if f.is_file())


@weave.type()
class SlackReadExportApi:
    data_dir: weave.Dir

    def channel_names(self):
        return (n.name for n in pathlib.Path(self.data_dir.path).glob("*"))

    def channel_path(self, channel_name):
        return pathlib.Path(self.data_dir.path) / channel_name

    def channel_export_size(self, channel_name: str):
        return dirsize(self.channel_path(channel_name))

    # Returns a dict in the slack export format. TODO
    def channel_messages(self, channel_name: str):
        files = self.channel_path(channel_name).glob("*.json")
        channel_messages = []
        for f in sorted(files):
            channel_messages += json.load(open(f))
        return channel_messages



import dataclasses
import typing
import weave

from .. import panel
from .. import graph
from .panel_group import PanelBankSectionConfig
from .bank import default_panel_bank_flow_section_config


PanelType = typing.TypeVar("PanelType")


@weave.type()
class EachConfig:
    pbConfig: PanelBankSectionConfig = dataclasses.field(
        default_factory=default_panel_bank_flow_section_config
    )
    panel: PanelType = dataclasses.field(default_factory=lambda: graph.VoidNode())  # type: ignore


@weave.type()
class Each(panel.Panel):
    id = "Each"
    input_node: graph.Node[list[typing.Any]]
    config: typing.Optional[EachConfig] = dataclasses.field(
        default_factory=lambda: None
    )

    def item_var(self):
        return graph.VarNode(self.input_node.type.object_type, "item")

    def __init__(self, input_node, vars=None, config=None, **options):
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            self.config = EachConfig()
            self.config.panel = self.item_var()



from __future__ import print_function
import dataclasses

import typing
import logging

from flask import current_app
from werkzeug.serving import make_server
import multiprocessing
import threading
import time
import requests
import traceback
import time
import pprint


from weave.language_features.tagging.tag_store import isolated_tagging_context
from . import value_or_error

from . import execute
from . import serialize
from . import storage
from . import context
from . import weave_types
from . import engine_trace
from . import logs
from . import wandb_api
from . import util
from . import graph
from .language_features.tagging import tag_store


# A function to monkeypatch the request post method
# def patch_request_post():
#     r_post = requests.post

#     logging.info = lambda *args, **kwargs: None
#     logging.warn = lambda *args, **kwargs: None
#     logging.debug = lambda *args, **kwargs: None

#     def post(*args, **kwargs):
#         logging.critical(kwargs["json"]["query"].split(' ')[1])
#         return r_post(*args, **kwargs)

#     requests.post = post

# patch_request_post()

PROFILE = False

OptionalAuthType = typing.Optional[
    typing.Union[typing.Tuple[str, str], requests.models.HTTPBasicAuth]
]

logger = logging.getLogger("root")


@dataclasses.dataclass
class HandleRequestResponse:
    results: value_or_error.ValueOrErrors[dict[str, typing.Any]]
    nodes: value_or_error.ValueOrErrors[graph.Node]


def handle_request(
    request, deref=False, serialize_fn=storage.to_python
) -> HandleRequestResponse:
    start_time = time.time()
    tracer = engine_trace.tracer()
    # nodes = [graph.Node.node_from_json(n) for n in request["graphs"]]
    with tracer.trace("request:deserialize"):
        nodes = serialize.deserialize(request["graphs"])

    with tracer.trace("request:execute"):
        with execute.top_level_stats() as stats:
            with context.execution_client():
                result = nodes.batch_map(execute.execute_nodes)
    with tracer.trace("request:deref"):
        if deref:
            result = result.zip(nodes).safe_map(
                lambda t: t[0]
                if isinstance(t[1].type, weave_types.RefType)
                else storage.deref(t[0])
            )

    # print("Server request %s (%0.5fs): %s..." % (start_time,
    #                                              time.time() - start_time, [n.from_op.name for n in nodes[:3]]))

    logging.info("FINAL STATS\n%s" % pprint.pformat(stats.op_summary()))

    # Forces output to be untagged
    with tracer.trace("serialize_response"):
        with isolated_tagging_context():
            with wandb_api.from_environment():
                result = result.safe_map(serialize_fn)

    logger.info("Server request done in: %ss" % (time.time() - start_time))
    tag_store.clear_tag_store()
    return HandleRequestResponse(result, nodes)


class SubprocessServer(multiprocessing.Process):
    def __init__(self, req_queue, resp_queue):
        multiprocessing.Process.__init__(self)
        self.req_queue = req_queue
        self.resp_queue = resp_queue

    def run(self):
        while True:
            req = self.req_queue.get()
            try:
                resp = handle_request(req).results.unwrap()
                self.resp_queue.put(resp)
            except:
                print("Weave subprocess server error")
                traceback.print_exc()
                self.resp_queue.put(Exception("Caught exception in sub-process server"))
                break

    def shutdown(self):
        self.kill()


class SubprocessServerClient:
    def __init__(self):
        self.req_queue = multiprocessing.Queue()
        self.resp_queue = multiprocessing.Queue()
        self.server_proc = SubprocessServer(self.req_queue, self.resp_queue)
        self.server_proc.start()

    def shutdown(self):
        self.server_proc.shutdown()

    def execute(self, nodes, no_cache=False):
        self.req_queue.put({"graphs": serialize.serialize(nodes)})
        response = self.resp_queue.get()
        deserialized = [storage.from_python(r) for r in response]
        return [storage.deref(r) for r in deserialized]


class InProcessServer(object):
    def __init__(self):
        pass

    def execute(self, nodes, no_cache=False):
        return execute.execute_nodes(nodes, no_cache=no_cache).unwrap()


class HttpServerClient(object):
    def __init__(self, url, emulate_weavejs=False, auth: OptionalAuthType = None):
        """Constructor.

        Args:
            url: The server base url
            emulate_weavejs: For testing only, should not be used from user code.
            auth (optional): auth argument to `requests.post`
        """
        self.url = url
        self.emulate_weavejs = emulate_weavejs
        self.execute_endpoint = "/__weave/execute/v2"
        self.auth = auth
        if emulate_weavejs:
            self.execute_endpoint = "/__weave/execute"

    def execute(self, nodes, no_cache=False):
        serialized = serialize.serialize(nodes)
        r = requests.post(
            self.url + self.execute_endpoint,
            json={"graphs": serialized},
            auth=self.auth,
        )
        r.raise_for_status()

        response = r.json()["data"]

        if self.emulate_weavejs:
            # When emulating weavejs, just return the server's json response.
            return response

        deserialized = [storage.from_python(r) for r in response]
        return [storage.deref(r) for r in deserialized]


_REQUESTED_SERVER_LOG_LEVEL: typing.Optional[int] = None


class HttpServer(threading.Thread):
    def __init__(self, port=0, host="localhost"):
        from . import weave_server

        self.host = host

        app = weave_server.app
        threading.Thread.__init__(self, daemon=True)
        self.srv = make_server(host, port, app, threaded=False)

        # if the passed port is zero then a randomly allocated port will be used. this
        # gets the value of the port that was assigned.
        self.port = self.srv.socket.getsockname()[1]

    @property
    def name(self):
        return f"Weave Port: {self.port}"

    def run(self):
        if _REQUESTED_SERVER_LOG_LEVEL is None:
            capture_weave_server_logs(logging.ERROR)

        # The werkzeug logger does not exist at import time, so we can't just
        # iterate through all existing loggers and set their levels. Fetching it
        # explicitly here creates it.
        # Without these two lines, request logs will always be printend. But by
        # default we don't want that.
        log = logging.getLogger("werkzeug")
        if _REQUESTED_SERVER_LOG_LEVEL is not None:
            log.setLevel(_REQUESTED_SERVER_LOG_LEVEL)

        self.srv.serve_forever()

    def shutdown(self):
        self.srv.shutdown()

    @property
    def url(self):
        if util.is_colab():
            url = f"https://{self.host}"
        else:
            url = f"http://{self.host}"
        if self.port is not None:
            url = f"{url}:{self.port}"
        return url


def capture_weave_server_logs(log_level: int = logging.INFO):
    global _REQUESTED_SERVER_LOG_LEVEL
    _REQUESTED_SERVER_LOG_LEVEL = log_level

    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)

    console_log_settings: typing.Optional[logs.LogSettings] = None
    if not util.is_notebook() or util.parse_boolean_env_var(
        "WEAVE_SERVER_FORCE_HTTP_SERVER_CONSOLE_LOGS"
    ):
        console_log_settings = logs.LogSettings(logs.LogFormat.PRETTY, level=None)

    logs.enable_stream_logging(
        root_logger,
        wsgi_stream_settings=console_log_settings,
        pid_logfile_settings=logs.LogSettings(logs.LogFormat.PRETTY, logging.INFO),
    )



import typing
import weakref
import contextlib
import contextvars
import collections

from . import uris
from . import box
from . import weave_types as types
from . import object_context

# We store Refs here if we can't attach them directly to the object
REFS: weakref.WeakValueDictionary[int, "Ref"] = weakref.WeakValueDictionary()

if typing.TYPE_CHECKING:
    from . import weave_types as types


class Ref:
    _obj: typing.Any
    _type: typing.Optional["types.Type"]
    extra: typing.Optional[list[str]]

    def __init__(
        self,
        obj: typing.Optional[typing.Any] = None,
        type: typing.Optional["types.Type"] = None,
        extra: typing.Optional[list[str]] = None,
    ):
        self._type = type
        self.extra = extra
        if obj is not None and type is not None and type.name != "tagged":
            obj = box.box(obj)
            _put_ref(obj, self)
        self._obj = obj

    @property
    def obj(self) -> typing.Any:
        obj_ctx = object_context.get_object_context()
        if obj_ctx:
            val = obj_ctx.lookup_ref_val(str(self.initial_uri))
            if not isinstance(val, object_context.NoResult):
                return val

        if self._obj is not None:
            return self._obj

        if not self.is_saved:
            return None

        obj = self._get()

        obj = box.box(obj)
        if self.type.name != "tagged":
            _put_ref(obj, self)
        self._obj = obj

        if obj_ctx is not None:
            obj_ctx.add_ref(str(self.initial_uri), obj, self.type)

        return obj

    @property
    def type(self) -> "types.Type":
        raise NotImplementedError

    @classmethod
    def from_str(cls, s: str) -> "Ref":
        return uris.WeaveURI.parse(s).to_ref()

    @property
    def initial_uri(self) -> str:
        raise NotImplementedError

    @property
    def uri(self) -> str:
        raise NotImplementedError

    @property
    def is_saved(self) -> bool:
        raise NotImplementedError

    def get(self) -> typing.Any:
        return self.obj

    def _get(self) -> typing.Any:
        raise NotImplementedError

    def __str__(self) -> str:
        return str(self.uri)


def get_ref(obj: typing.Any) -> typing.Optional[Ref]:
    if isinstance(obj, Ref):
        return obj
    if hasattr(obj, "_ref"):
        return obj._ref
    try:
        if id(obj) in REFS:
            return REFS[id(obj)]
    except TypeError:
        pass
    return None


# TODO: this cannot be used on tagged objects, since we don't know
# whether obj is intended to be the tagged or untagged version
# So for now we never call it on tagged objects. But this probably
# breaks some behaviors (like automatic cross-artifact references
# for tagged objects)
def _put_ref(obj: typing.Any, ref: Ref) -> None:
    try:
        obj._ref = ref
    except (AttributeError, ValueError):
        if isinstance(obj, (int, float, str, list, dict, set)):
            return
        REFS[id(obj)] = ref


def clear_ref(obj: typing.Any) -> None:
    try:
        delattr(obj, "_ref")
    except AttributeError:
        pass
    if id(obj) in REFS:
        REFS.pop(id(obj))


def deref(ref: Ref) -> typing.Any:
    if isinstance(ref, Ref):
        return ref.get()
    return ref


types.RefType.instance_classes = Ref



import dataclasses
import typing

import weave
from weave import weave_internal
from weave import codify
from weave import codifiable_value_mixin
from weave.graph import ConstNode, Node, OutputNode, VoidNode
from .. import panel
from . import table_state

from .panel_query import Query


@weave.type("tablePanelConfig")
class TableConfig:
    tableState: table_state.TableState
    rowSize: int = dataclasses.field(default_factory=lambda: 1)
    pinnedRows: dict[str, list[int]] = dataclasses.field(default_factory=dict)
    activeRowForGrouping: dict[str, int] = dataclasses.field(default_factory=dict)


class ColumnDef(typing.TypedDict):
    columnName: str
    columnSelectFunction: weave.graph.Node


@dataclasses.dataclass
class TableColumn:
    select: typing.Callable
    groupby: bool = dataclasses.field(default=False)
    name: str = dataclasses.field(default_factory=lambda: "")


@weave.type("tablePanel")
class Table(panel.Panel, codifiable_value_mixin.CodifiableValueMixin):
    id = "table"
    input_node: weave.Node[list[typing.Any]]
    config: typing.Optional[TableConfig] = dataclasses.field(
        default_factory=lambda: None
    )

    def __init__(self, input_node, vars=None, config=None, **options):
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            table = table_state.TableState(self.input_node)
            self.config = TableConfig(table)

            if "columns" in options:
                for col_ndx, column_expr in enumerate(options["columns"]):
                    if isinstance(column_expr, TableColumn):
                        table.add_column(
                            column_expr.select,
                            column_expr.name,
                            groupby=column_expr.groupby,
                        )
                    else:
                        table.add_column(column_expr)

    def get_final_named_select_functions(self) -> dict[str, ColumnDef]:
        if not self.config:
            return {}
        else:
            result = {}
            existing_names: typing.Set[str] = set()
            group_columns = self.config.tableState.groupBy
            use_order = group_columns + [
                key for key in self.config.tableState.order if key not in group_columns
            ]
            for internal_col_key in use_order:
                fn_node = self.config.tableState.columnSelectFunctions[internal_col_key]
                if fn_node.type != weave.types.Invalid():
                    name = self.config.tableState.columnNames[internal_col_key]
                    if name == "":
                        name = "c_" + str(len(existing_names))
                    added_suffix = 1
                    _name = name
                    while _name in result:
                        _name = f"{name}_{added_suffix}"
                        added_suffix += 1
                    name = _name

                    existing_names.add(name)
                    result[internal_col_key] = ColumnDef(
                        columnName=name,
                        columnSelectFunction=fn_node,
                    )
        return result

    def to_code(self) -> typing.Optional[str]:
        field_vals: list[tuple[str, str]] = []
        if self.config is not None:
            ts = self.config.tableState
            if ts.autoColumns:
                return None
            if not isinstance(ts.preFilterFunction, VoidNode):
                return None
            if ts.groupBy != []:
                return None
            if ts.sort != []:
                return None

            default_column_panel_def = table_state.PanelDef("", {}, None)
            default_column_panel_def_as_dict = {
                "panelId": "",
                "panelConfig": None,
                "originalKey": "",
            }

            code_cols = []
            for col_id in ts.order:
                if ts.columnNames[col_id] != "":
                    return None
                if (
                    ts.columns[col_id] != default_column_panel_def
                    and ts.columns[col_id] != default_column_panel_def_as_dict
                ):
                    return None
                code_cols.append(
                    codify.lambda_wrapped_object_to_code_no_format(
                        ts.columnSelectFunctions[col_id], ["row"]
                    )
                )
            if len(code_cols) > 0:
                field_vals.append(("columns", "[" + ",".join(code_cols) + ",]"))

        param_str = ""
        if len(field_vals) > 0:
            param_str = (
                ",".join([f_name + "=" + f_val for f_name, f_val in field_vals]) + ","
            )
        return f"""weave.panels.panel_table.Table({codify.object_to_code_no_format(self.input_node)}, {param_str})"""

    def add_column(self, select_expr, name=None):
        self.config.tableState.add_column(select_expr, name)


def _get_composite_group_key(self: typing.Union[Table, Query]) -> str:
    if self.config is None:
        return ""
    group_by_keys = self.config.tableState.groupBy
    group_by_keys = group_by_keys or []
    return ",".join(group_by_keys)


# TODO: preserve arrow
def _get_pinned_node(self: typing.Union[Table, Query], data_or_rows_node: Node) -> Node:
    if self.config is None:
        return weave.ops.make_list()

    composite_group_key = _get_composite_group_key(self)
    pinned_data = self.config.pinnedRows.get(composite_group_key)
    if pinned_data is None or len(pinned_data) == 0:
        return weave.ops.make_list()

    return weave.ops.make_list(
        **{
            f"v_{pin_ndx}": OutputNode(
                data_or_rows_node.type,
                "index",
                {
                    "arr": data_or_rows_node,
                    "index": ConstNode(weave.types.Int(), row_ndx),
                },
            )
            for pin_ndx, row_ndx in enumerate(pinned_data)
        }
    )


def _get_active_node(self: Table, data_or_rows_node: Node) -> Node:
    composite_group_key = _get_composite_group_key(self)
    if self.config is None or self.config.activeRowForGrouping is None:
        active_index = 0
    else:
        active_index = self.config.activeRowForGrouping.get(composite_group_key, 0)

    return OutputNode(
        data_or_rows_node.type,
        "list-__getitem__",
        {
            "arr": data_or_rows_node,
            "index": ConstNode(weave.types.Int(), active_index),
        },
    )


# TODO: preserve arrow for empty list
def _get_rows_node(self: Table) -> Node:
    # Apply Filters
    data_node = self.input_node
    if (
        self.config
        and self.config.tableState.preFilterFunction is not None
        and self.config.tableState.preFilterFunction.type != weave.types.Invalid()
    ):
        data_node = weave.ops.List.filter(
            data_node,
            lambda row, index: weave_internal.call_fn(
                self.config.tableState.preFilterFunction, {"row": row, "index": index}
            ),
        )

    columns = self.get_final_named_select_functions()

    # Apply Grouping
    group_ids: typing.Set[str] = set()
    if self.config and self.config.tableState.groupBy:
        group_ids = set(self.config.tableState.groupBy)
        data_node = weave.ops.List.groupby(
            data_node,
            lambda row: weave.ops.dict_(
                **{
                    columns[col_id]["columnName"]: weave_internal.call_fn(
                        columns[col_id]["columnSelectFunction"],
                        {
                            "row": row,
                        },
                    )
                    for col_id in self.config.tableState.groupBy
                }
            ),
        )

    # Apply Selection
    data_node = weave.ops.List.map(
        data_node,
        lambda row, index: weave.ops.dict_(
            **{
                col_def["columnName"]: (
                    weave_internal.call_fn(
                        col_def["columnSelectFunction"], {"row": row, "index": index}
                    )
                    if col_id not in group_ids
                    else
                    # Here, we materialize the groupkey tags as columns. This is nicer for downstream
                    # consumers, to be able to directly fetch the col by name, even if it
                    # was the grouping column. I (Tim) took this design liberty, easy to change.
                    row.groupkey()[col_def["columnName"]]
                )
                for col_id, col_def in columns.items()
            }
        ),
    )

    # Apply Sorting
    if self.config and self.config.tableState.sort:
        sort_defs = self.config.tableState.sort

        def make_sort_fn(sort_def, row_node):
            col_name = columns[sort_def["columnId"]]["columnName"]
            return row_node[col_name]
            # If we don't materialize the group column above, we need to do
            # something like this
            # if sort_def['columnId'] in group_ids:
            #     return row_node.groupkey()[col_name]
            # else:
            #     return row_node[col_name]

        data_node = weave.ops.List.sort(
            data_node,
            lambda row: weave.ops.make_list(
                **{
                    f"{sort_ndx}": make_sort_fn(sort_def, row)
                    for sort_ndx, sort_def in enumerate(sort_defs)
                }
            ),
            [sort_def["dir"] for sort_def in sort_defs],
        )

    return data_node


def _get_row_type(self: Table) -> weave.types.Type:
    columns = self.get_final_named_select_functions()
    inner_type = weave.types.TypedDict(
        {
            col_def["columnName"]: col_def["columnSelectFunction"].type
            for col_def in columns.values()
        }
    )

    # Since we are executing the node, tags don't come out!
    # if self.config and self.config.tableState.groupBy:
    #     inner_type = tagged_value_type.TaggedValueType(weave.types.TypedDict({
    #         "groupKey": weave.types.TypedDict({
    #             columns[key]['columnName']: columns[key]["columnSelectFunction"].type
    #             for key in self.config.tableState.groupBy
    #         })
    #     }), inner_type)

    return inner_type


# TODO: preserve arrow
@weave.op(name="panel_table-rows_refine", hidden=True)
def rows_refine(self: Table) -> weave.types.Type:
    return weave.types.List(_get_row_type(self))


@weave.op(name="panel_table-rows_single_refine", hidden=True)
def rows_single_refine(self: Table) -> weave.types.Type:
    return weave.types.optional(_get_row_type(self))


@weave.op(name="panel_table-data_refine", hidden=True)
def data_refine(self: Table) -> weave.types.Type:
    return self.input_node.type


@weave.op(name="panel_table-data_single_refine", hidden=True)
def data_single_refine(self: Table) -> weave.types.Type:
    if not hasattr(self.input_node.type, "object_type"):
        return weave.types.Any()
    # input_node.type should be FunctionType (since its a Node)
    # but for some reason its not.
    # TODO: fix
    return weave.types.optional(self.input_node.type.object_type)  # type: ignore


# TODO: keep type in arrow
@weave.op(
    name="panel_table-all_rows",
    output_type=weave.types.List(weave.types.TypedDict({})),
    refine_output_type=rows_refine,
)
def rows(self: Table):
    rows_node = _get_rows_node(self)
    return rows_node


# Defined on both Table and Query
@weave.op(
    name="panel_table-pinned_data",
    # we define output_type for pyton and refine_output_type for JS
    output_type=lambda inputs: inputs["self"].input_node.output_type,
    refine_output_type=data_refine,
)
def pinned_data(self: typing.Union[Table, Query]):
    pinned_data_node = _get_pinned_node(self, self.input_node)
    return weave.use(pinned_data_node)


# TODO: preserve arrow
@weave.op(
    name="panel_table-pinned_rows",
    output_type=weave.types.List(weave.types.TypedDict({})),
    refine_output_type=rows_refine,
)
def pinned_rows(self: Table):
    rows_node = _get_rows_node(self)
    pinned_data_node = _get_pinned_node(self, rows_node)
    return weave.use(pinned_data_node)


@weave.op(
    name="panel_table-active_data",
    # We can't use the refine_output_type here, because it will become
    # non-weavifiable and not sent over the wire.
    # output_type=lambda inputs: inputs['self'].input_node.output_type.object_type,
    refine_output_type=data_single_refine,
)
def active_data(self: Table) -> typing.Optional[typing.Any]:
    data_node = _get_active_node(self, self.input_node)
    return data_node


@weave.op(
    name="panel_table-active_row",
    output_type=weave.types.optional(weave.types.TypedDict({})),
    refine_output_type=rows_single_refine,
)
def active_row(self: Table):
    rows_node = _get_rows_node(self)
    data_node = _get_active_node(self, rows_node)
    return weave.use(data_node)



from ..compile_domain import wb_gql_op_plugin
from ..api import op
from .. import weave_types as types
from . import wb_domain_types as wdt
from ..language_features.tagging.make_tag_getter_op import make_tag_getter_op
from .wandb_domain_gql import (
    gql_prop_op,
    gql_direct_edge_op,
    gql_connection_op,
    gql_root_op,
)

# Section 1/6: Tag Getters
# None

# Section 2/6: Root Ops
entity = gql_root_op(
    "root-entity",
    "entity",
    wdt.EntityType,
    {
        "entityName": types.String(),
    },
    lambda inputs: f'name: {inputs["entityName"]}',
)


# Section 3/6: Attribute Getters
@op(name="entity-name")
def entity_name(entity: wdt.Entity) -> str:
    return entity.gql["name"]


gql_prop_op(
    "entity-isTeam",
    wdt.EntityType,
    "isTeam",
    types.Boolean(),
)

# Section 4/6: Direct Relationship Ops
gql_direct_edge_op(
    "entity-org",
    wdt.EntityType,
    "organization",
    wdt.OrgType,
)

# Section 5/6: Connection Ops
gql_connection_op(
    "entity-portfolios",
    wdt.EntityType,
    "artifactCollections",
    wdt.ArtifactCollectionType,
    {},
    lambda inputs: f"collectionTypes: [PORTFOLIO]",
)

gql_connection_op(
    "entity-projects",
    wdt.EntityType,
    "projects",
    wdt.ProjectType,
    {},
    # lambda inputs: f"first: 100",
)

# Section 6/6: Non Standard Business Logic Ops
@op(name="entity-link")
def entity_link(entity: wdt.Entity) -> wdt.Link:
    return wdt.Link(entity.gql["name"], f"/{entity.gql['name']}")



import inspect

from . import registry_mem
from . import op_def
from . import op_args
from . import derive_op
from . import weave_types as types
from . import pyfunc_type_util


def op(
    input_type=None,
    output_type=None,
    refine_output_type=None,
    name=None,
    setter=None,
    render_info=None,
    hidden=False,
    pure=True,
    returns_expansion_node=False,
    _op_def_class=op_def.OpDef,
    *,  # Marks the rest of the arguments as keyword-only.
    plugins=None,
    mutation=False,
):
    """Decorator for declaring an op.

    Decorated functions must be typed, either with Python types or by declaring
    input_type, output_type as arguments to op (Python types preferred).
    """

    def wrap(f):
        weave_input_type = pyfunc_type_util.determine_input_type(f, input_type)
        weave_output_type = pyfunc_type_util.determine_output_type(f, output_type)

        fq_op_name = name
        if fq_op_name is None:
            op_prefix = "op"

            # This would be a much nicer: automatically use first type name.
            # But what about Unions? I'm not going to solve in this PR.
            # TODO: implement this.
            # if (
            #     isinstance(weave_input_type, op_args.OpNamedArgs)
            #     and weave_input_type.arg_types
            # ):
            #     arg_type0 = next(iter(weave_input_type.arg_types.values()))
            #     if arg_type0 != types.UnknownType():
            #         op_prefix = arg_type0.name

            fq_op_name = f"{op_prefix}-{f.__name__}"

        op = _op_def_class(
            fq_op_name,
            weave_input_type,
            weave_output_type,
            f,
            refine_output_type=refine_output_type,
            setter=setter,
            render_info=render_info,
            hidden=hidden,
            pure=pure,
            plugins=plugins,
            mutation=mutation,
            returns_expansion_node=returns_expansion_node,
            _decl_locals=inspect.currentframe().f_back.f_locals,
        )

        op_version = registry_mem.memory_registry.register_op(op)

        # After we register the op, create any derived ops

        # If op.location is set, then its a custom (non-builtin op). We don't
        # derive custom ops for now, as the derive code doesn't do the right thing.
        # The op name is the location/uri for custom ops, and the derive code doesn't
        # fix that up. So we end up double registering ops in WeaveJS which breaks
        # everything.
        # TODO: fix so we get mappability for custom (ecosystem) ops!
        if op.location is None:
            derive_op.derive_ops(op)

        return op_version

    return wrap



import typing
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

from ..ops_primitives import list_ as primitive_list
from .. import weave_types as types
from ..language_features.tagging import tagged_value_type
from ..api import op
from .. import graph
from .. import engine_trace

from . import convert
from .arrow import (
    ArrowWeaveListType,
    safe_coalesce,
)
from .arrow_tags import pushdown_list_tags
from .list_ import ArrowWeaveList, make_vec_dict, make_vec_taggedvalue, awl_zip

tracer = engine_trace.tracer()  # type: ignore


def join_all_zip_impl_lambda(
    arrs: list[ArrowWeaveList], joinFn: graph.OutputNode, outer: bool
):
    # Filter Nones and pushdown tags
    arrs = [pushdown_list_tags(a) for a in arrs if a != None]

    # Get the join key for each list
    join_key_cols = [a.apply(joinFn).untagged() for a in arrs]

    return join_all_zip_impl_vec(arrs, join_key_cols, outer)


def join_all_zip_impl_vec(
    arrs: list[ArrowWeaveList], join_key_cols: list[ArrowWeaveList], outer: bool
):
    # Empty case
    if len(arrs) == 0:
        # TODO: I don't think the arrow type is correct here. Should turn validate on.
        return ArrowWeaveList(pa.array([]), types.TypedDict())

    # Do the join
    if outer:
        join_type = "full outer"
    else:
        join_type = "inner"
    aliases = [f"a{i}" for i in range(len(arrs))]
    join_key_col, join_result = join_all_impl(arrs, join_key_cols, aliases, join_type)

    all_keys = {}
    for a in arrs:
        all_keys.update(dict.fromkeys(a.keys()))

    zipped_cols = []
    for k in all_keys:
        key_cols = [new_arr.column(k) for new_arr in join_result]
        zipped_col = awl_zip(*key_cols)
        zipped_cols.append(zipped_col)

    return make_vec_taggedvalue(
        make_vec_dict(joinObj=join_key_col),
        make_vec_dict(**{k: v for k, v in zip(all_keys, zipped_cols)}),
    )


def join_all_impl(
    arrs: list[ArrowWeaveList],
    join_key_cols: list[ArrowWeaveList],
    aliases: list[str],
    join_type: str,
) -> tuple[ArrowWeaveList, list[ArrowWeaveList]]:
    # Ensure each join key col has the same type
    unified_join_key_cols = convert.unify_types(*join_key_cols)

    join_key_cols = unified_join_key_cols

    # Get the safe join keys
    safe_join_key_cols = [convert.to_compare_safe(a) for a in join_key_cols]

    tables: list[pa.Table] = []
    for i, (arr, safe_join_key_col) in enumerate(zip(arrs, safe_join_key_cols)):
        if pa.types.is_null(safe_join_key_col._arrow_data.type):
            # Special case the type of a null column. Arrow's won't join on it.
            # But nulls can be represented in any type, so we cast to int64. The
            # safe join column is not included in the final output, so we don't
            # need to worry about fixing the type later.
            safe_join_key_col._arrow_data = safe_join_key_col._arrow_data.cast("int64")
        tables.append(
            pa.Table.from_arrays(
                [safe_join_key_col._arrow_data, np.arange(len(arr), dtype="int64")],
                names=["join", f"index_t{i}"],
            ).filter(pc.invert(pc.is_null(safe_join_key_col._arrow_data)))
        )

    joined = tables[0]
    for other in tables[1:]:
        joined = joined.join(
            other,
            ["join"],
            join_type=join_type,
            use_threads=False,
            coalesce_keys=True,
        )

    index_cols = [joined[f"index_t{i}"].combine_chunks() for i in range(len(arrs))]

    final_join_key_col: ArrowWeaveList = ArrowWeaveList(
        safe_coalesce(
            *[
                a._arrow_data.take(index_col)
                for index_col, a in zip(index_cols, join_key_cols)
            ]
        ),
        join_key_cols[0].object_type,
        None,
    )

    new_arrs: list[ArrowWeaveList] = []
    for index_col, arr in zip(index_cols, arrs):
        new_arrs.append(
            ArrowWeaveList(
                arr._arrow_data.take(index_col),
                types.optional(arr.object_type),
                arr._artifact,
            )
        )

    return final_join_key_col, new_arrs


Obj1Type = typing.TypeVar("Obj1Type")
Obj2Type = typing.TypeVar("Obj2Type")


def join2_impl(
    arr1: ArrowWeaveList,
    arr2: ArrowWeaveList,
    join1Fn: graph.OutputNode,
    join2Fn: graph.OutputNode,
    alias1: typing.Optional[str] = None,
    alias2: typing.Optional[str] = None,
    leftOuter: bool = False,
    rightOuter: bool = False,
):
    if alias1 == None:
        alias1 = "a1"
    if alias2 == None:
        alias2 = "a2"

    arr1_raw_join_col = arr1.apply(join1Fn).untagged()
    arr2_raw_join_col = arr2.apply(join2Fn).untagged()

    if leftOuter and rightOuter:
        join_type = "full outer"
    elif leftOuter:
        join_type = "left outer"
    elif rightOuter:
        join_type = "right outer"
    else:
        join_type = "inner"

    join_key_col, [new_arr1, new_arr2] = join_all_impl(
        [arr1, arr2],
        [arr1_raw_join_col, arr2_raw_join_col],
        [alias1, alias2],  # type: ignore
        join_type,
    )

    return make_vec_taggedvalue(
        make_vec_dict(joinObj=join_key_col),
        make_vec_dict(**{alias1: new_arr1, alias2: new_arr2}),  # type: ignore
    )


def _join_all_output_type(input_types):
    return _joined_all_output_type_of_arrs_type(
        input_types["arrs"].object_type.object_type, input_types["joinFn"].output_type
    )


def _joined_all_output_type_of_arrs_type(
    arr_obj_type: types.TypedDict, join_fn_output_type: types.Type
) -> types.Type:
    inner_type = _joined_all_output_type_inner_type(
        arr_obj_type,
    )
    tag_type = _joined_all_output_type_tag_type(join_fn_output_type)
    tagged_type = tagged_value_type.TaggedValueType(tag_type, inner_type)
    return ArrowWeaveListType(tagged_type)


def _joined_all_output_type_inner_type(
    arr_obj_type: types.TypedDict,
) -> types.TypedDict:
    arr_prop_types = arr_obj_type.property_types
    prop_types: dict[str, types.Type] = {}
    for k in arr_prop_types.keys():
        prop_types[k] = types.List(types.optional(arr_prop_types[k]))
    return types.TypedDict(prop_types)


def _joined_all_output_type_tag_type(
    join_fn_output_type: types.Type,
) -> types.TypedDict:
    return types.TypedDict({"joinObj": join_fn_output_type})


@op(
    name="ArrowWeaveList-joinAll",
    input_type={
        "arrs": types.List(types.optional(ArrowWeaveListType(types.TypedDict({})))),
        "joinFn": lambda input_types: types.Function(
            {"row": input_types["arrs"].object_type.object_type}, types.Any()
        ),
    },
    output_type=_join_all_output_type,
)
def join_all(arrs, joinFn, outer: bool):
    return join_all_zip_impl_lambda(arrs, joinFn, outer)


def _join_2_output_type(input_types):
    return ArrowWeaveListType(primitive_list._join_2_output_row_type(input_types))


@op(
    name="ArrowWeaveList-join",
    input_type={
        "arr1": ArrowWeaveListType(types.TypedDict({})),
        "arr2": ArrowWeaveListType(types.TypedDict({})),
        "join1Fn": lambda input_types: types.Function(
            {"row": input_types["arr1"].object_type}, types.Any()
        ),
        "join2Fn": lambda input_types: types.Function(
            {"row": input_types["arr2"].object_type}, types.Any()
        ),
        "alias1": types.String(),
        "alias2": types.String(),
        "leftOuter": types.Boolean(),
        "rightOuter": types.Boolean(),
    },
    output_type=_join_2_output_type,
)
def join_2(arr1, arr2, join1Fn, join2Fn, alias1, alias2, leftOuter, rightOuter):
    return join2_impl(
        arr1, arr2, join1Fn, join2Fn, alias1, alias2, leftOuter, rightOuter
    )
